from collections import namedtuple


def pRule(peg, name, pf):
    peg[name] = pf

# Generator


def match_empty(px): return True


def pEmpty():
    return match_empty


def pFail():
    return lambda px: False


def match_any(px):
    if px.pos < px.epos:
        px.pos += 1
        return True
    return False


def pAny():
    return match_any


CharCache = {
    '': match_empty
}


def pChar(text):
    if text in CharCache:
        return CharCache[text]
    clen = len(text)

    def match_char(px):
        if px.inputs.startswith(text, px.pos):
            px.pos += clen
            return True
        return False
    CharCache[text] = match_char
    return match_char

# Range


BitmapCache = {}


def unique_range(chars, ranges, memo=None):
    cs = 0
    for c in chars:
        cs |= 1 << ord(c)
    r = ranges
    while len(r) > 1:
        for c in range(ord(r[0]), ord(r[1])+1):
            cs |= 1 << c
        r = r[2:]
    if memo is not None:
        if cs in memo:
            return memo[cs]
        memo[cs] = cs
    return cs


'''
def minimum_range(chars, ranges):
    cs = 0xffff
    for c in chars:
        cs = min(cs, ord(c))
    r = ranges
    while len(r) > 1:
        cs = min(cs, ord(r[0]))
        cs = min(cs, ord(r[1]))
        r = r[2:]
    return cs




def bitmap(chars, ranges):
    key = (chars, ranges)
    if key in BitmapCache:
        return BitmapCache[key]
    offset = minimum_range(chars, ranges)
    bitset = unique_range(chars, ranges) >> offset
    BitmapCache[key] = (bitset, offset)
    return BitmapCache[key]


def pRange(chars, ranges):
    bitset, offset = bitmap(chars, ranges)

    def match_bitset(px):
        if px.pos < px.epos:
            shift = ord(px.inputs[px.pos]) - offset
            if shift >= 0 and (bitset & (1 << shift)) != 0:
                px.pos += 1
                return True
        return False
    return match_bitset
'''

# NewRange


def tochars(chars, ranges):
    cs = set(list(chars))
    rs = ranges
    while len(rs) > 0:
        r = range(ord(rs[0]), ord(rs[1])+1)
        cs |= set(map(chr, r))
        rs = rs[2:]
    return ''.join(sorted(cs))


def make_bitset(chars, ranges, NBITS=8):
    cmin = min(map(ord, list(chars+ranges)))-1
    bitmap = [0] * ((max(map(ord, list(chars+ranges))) - cmin) // NBITS + 1)
    for c in chars:
        n = (ord(c) - cmin) // NBITS
        mask = 1 << (ord(c) - cmin) % NBITS
        bitmap[n] |= mask
    r = ranges
    while len(r) > 1:
        for c in range(ord(r[0]), ord(r[1])+1):
            n = (c - cmin) // NBITS
            mask = 1 << (c - cmin) % NBITS
            bitmap[n] |= mask
        r = r[2:]
    return bitmap, cmin, len(bitmap)*NBITS


def range_pattern(chars, ranges):
    key = tochars(chars, ranges)
    if len(key) < 1000:
        return key
    if key in BitmapCache:
        return BitmapCache[key]
    BitmapCache[key] = make_bitset(chars, ranges)
    return BitmapCache[key]


def match_chars(px, chars):
    return px.pos < px.epos and chars.find(px.inputs[px.pos]) != -1


def match_bitset(px, bitset, offset, maxlen):
    if px.pos < px.epos:
        c = ord(px.inputs[px.pos]) - offset
        if c < 0 or c >= maxlen:
            return False
        mask = 1 << c % 8
        return bitset[c//8] & mask == mask
    return False


def pAndRange(chars, ranges):
    pat = range_pattern(chars, ranges)
    if isinstance(pat, str):
        def match_chars(px):
            return px.pos < px.epos and pat.find(px.inputs[px.pos]) != -1
        return match_chars

    bitset, offset, maxlen = pat

    def match_ranges(px):
        if px.pos < px.epos:
            c = ord(px.inputs[px.pos]) - offset
            if c < 0 or c >= maxlen:
                return False
            mask = 1 << (c % 8)
            return bitset[c//8] & mask == mask
        return False
    return match_ranges


def pRange(chars, ranges):
    pf = pAndRange(chars, ranges)

    def match_ranges(px):
        if pf(px):
            px.pos += 1
            return True
        return False
    return match_ranges


def pAnd(pf):
    def match_and(px):
        pos = px.pos
        if pf(px):
            px.headpos = max(px.pos, px.headpos)
            px.pos = pos
            return True
        return False
    return match_and


def pNot(pf):
    def match_not(px):
        pos = px.pos
        ptree = px.ptree
        if not pf(px):
            px.headpos = max(px.pos, px.headpos)
            px.pos = pos
            px.ptree = ptree
            return True
        return False
    return match_not


def pMany(pf):
    def match_many(px):
        pos = px.pos
        ptree = px.ptree
        while pf(px) and pos < px.pos:
            pos = px.pos
            ptree = px.ptree
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ptree = ptree
        return True
    return match_many


def pOneMany(pf):
    def match_OneMany(px):
        if pf(px):
            pos = px.pos
            ptree = px.ptree
            while pf(px) and pos < px.pos:
                pos = px.pos
                ptree = px.ptree
            px.headpos = max(px.pos, px.headpos)
            px.pos = pos
            px.ptree = ptree
            return True
        return False
    return match_OneMany


def pOption(pf):
    def match_option(px):
        pos = px.pos
        ptree = px.ptree
        if not pf(px):
            px.headpos = max(px.pos, px.headpos)
            px.pos = pos
            px.ptree = ptree
        return True
    return match_option

# Seq


def pSeq2(pf, pf2):
    def match_seq2(px):
        return pf(px) and pf2(px)
    return match_seq2


def pSeq3(pf, pf2, pf3):
    def match_seq3(px):
        return pf(px) and pf2(px) and pf3(px)
    return match_seq3


def pSeq4(pf, pf2, pf3, pf4):
    def match_seq4(px):
        return pf(px) and pf2(px) and pf3(px) and pf4(px)
    return match_seq4


def pSeq(*pfs):
    def match_seq(px):
        for pf in pfs:
            if not pf(px):
                return False
        return True
    return match_seq

# Ore


def pOre2(pf, pf2):
    def match_ore2(px):
        pos = px.pos
        ptree = px.ptree
        if pf(px):
            return True
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ptree = ptree
        return pf2(px)
    return match_ore2


def pOre3(pf, pf2, pf3):
    def match_ore3(px):
        pos = px.pos
        ptree = px.ptree
        if pf(px):
            return True
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ptree = ptree
        if pf2(px):
            return True
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ptree = ptree
        return pf3(px)
    return match_ore3


def pOre4(pf, pf2, pf3, pf4):
    def match_ore4(px):
        pos = px.pos
        ptree = px.ptree
        if pf(px):
            return True
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ptree = ptree
        if pf2(px):
            return True
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ptree = ptree
        if pf3(px):
            return True
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ptree = ptree
        return pf4(px)
    return match_ore4


def pOre(*pfs):
    def match_ore(px):
        pos = px.pos
        ptree = px.ptree
        for pf in pfs:
            if pf(px):
                return True
            px.headpos = max(px.pos, px.headpos)
            px.pos = pos
            px.ptree = ptree
        return False
    return match_ore


def make_trie(dic):
    if '' in dic or len(dic) < 10:
        return dic
    d = {}
    for s in dic:
        s0, s = s[0], s[1:]
        if s0 in d:
            ss = d[s0]
            if not s in ss:
                ss.append(s)
        else:
            d[s0] = [s]
    for key in d:
        d[key] = make_trie(d[key])
    return d


def match_trie(px, d):
    if px.pos >= px.epos:
        return False
    if isinstance(d, dict):
        c = px.inputs[px.pos]
        if c in d:
            px.pos += 1
            return match_trie(px, d[c])
        return False
    pos = px.pos
    inputs = px.inputs
    for s in d:
        if inputs.startswith(s, pos):
            px.pos += len(s)
            return True
    return False


def pDict(words):
    if isinstance(words, str):
        words = words.split(' ')
    dic = make_trie(words)
    return lambda px: match_trie(px, dic)


# def pRef(generated, uname):
#     if uname not in generated:
#         fs = None

#         def match_deref(px):
#             nonlocal fs
#             if fs is None:
#                 fs = generated[uname]
#             return fs(px)
#         generated[uname] = match_deref
#     return generated[uname]


def pRef(generated, uname):
    if uname not in generated:
        generated[uname] = lambda px: generated[uname](px)
    return generated[uname]


class PMemo(object):
    __slots__ = ['key', 'pos', 'treeState', 'ptree', 'prev', 'result']

    def __init__(self):
        self.key = -1
        self.pos = 0
        self.treeState = False
        self.ptree = None
        self.prev = None
        self.result = False


def pMemo(fs, mp, mpsize):
    disabled = False
    hit = 0
    miss = 0

    def match_memo(px):
        nonlocal disabled, hit, miss
        if disabled:
            return fs(px)
        key = (mpsize * px.pos) + mp
        m = px.memo[key % 1789]
        if m.key == key:
            if m.treeState:
                if m.prev == px.ptree:
                    px.pos = m.pos
                    px.ptree = m.ptree
                    hit += 1
                    return m.result
            else:
                px.pos = m.pos
                hit += 1
                return m.result
        prev = px.ptree
        m.result = fs(px)
        m.pos = px.pos
        m.key = key
        if m.result and prev != px.ptree:
            m.treeState = True
            m.prev = prev
            m.ptree = px.ptree
        else:
            m.treeState = False
        miss += 1
        if miss % 100 == 0:
            if hit / miss < 5:
                disabled = True
        return m.result
    return match_memo


def pMemoDebug(name, fs, mp, mps):
    disabled = False
    hit = 0
    miss = 0
    mpsize = len(mps)

    def match_memo(px):
        nonlocal disabled, hit, miss
        if disabled:
            return fs(px)
        key = (mpsize * px.pos) + mp
        m = px.memo[key % 1789]
        if m.key == key:
            if m.treeState:
                if m.prev == px.ptree:
                    px.pos = m.pos
                    px.ptree = m.ptree
                    hit += 1
                    return m.result
            else:
                px.pos = m.pos
                hit += 1
                return m.result
        prev = px.ptree
        m.result = fs(px)
        m.pos = px.pos
        m.key = key
        if m.result and prev != px.ptree:
            m.treeState = True
            m.prev = prev
            m.ptree = px.ptree
        else:
            m.treeState = False
        miss += 1
        if miss % 100 == 0:
            if hit / miss < 5:
                mps.remove(name)
                print('enabled', mps)
                disabled = True
        return m.result
    return match_memo


# Tree Construction


class PTree(object):
    __slots__ = ['prev', 'tag', 'spos', 'epos', 'child']

    def __init__(self, prev, tag, spos, epos, child):
        self.prev = prev
        self.tag = tag
        self.spos = spos
        self.epos = epos
        self.child = child

    def isEdge(self):
        return self.epos < 0

    def dump(self, inputs):
        sb = []
        if self.prev is not None:
            sb.append(self.prev.dump(inputs))
            sb.append(',')
        sb.append(f'{{#{self.tag} ')
        if self.child is None:
            sb.append(repr(inputs[self.spos:self.epos]))
        else:
            sb.append(self.child.dump(inputs))
        sb.append('}')
        return ''.join(sb)


def pNode(pf, tag, shift):
    def make_tree(px):
        pos = px.pos
        prev = px.ptree
        px.ptree = None
        if pf(px):
            px.ptree = PTree(prev, tag, pos+shift, px.pos, px.ptree)
            return True
        return False
    return make_tree


def pEdge(edge, pf, shift=0):
    def match_edge(px):
        pos = px.pos
        prev = px.ptree
        px.ptree = None
        if pf(px):
            if px.ptree is None:
                px.ptree = PTree(None, '', pos+shift, px.pos, px.ptree)
            px.ptree = PTree(prev, edge, -1, -1, px.ptree)
            return True
        return False
    return match_edge


def popPTree(px):
    pt = px.ptree
    if pt is None:
        return px.pos, None
    if pt.prev is None:
        px.ptree = pt
        return pt.spos, None
    px.ptree = PTree(None, pt.tag, pt.spos, pt.epos, pt.child)
    return pt.spos, pt.prev


def pFold(edge, pf, tag, shift):
    if edge == '':
        def match_fold(px):
            pos, prev = popPTree(px)
            if pf(px):
                px.ptree = PTree(prev, tag, pos, px.pos, px.ptree)
                return True
            return False
        return match_fold
    else:
        def match_fold2(px):
            pos, prev = popPTree(px)
            px.ptree = PTree(None, edge, -1, -1, px.ptree)
            if pf(px):
                px.ptree = PTree(prev, tag, pos, px.pos, px.ptree)
                return True
            return False
        return match_fold2


def pAbs(pf):
    def match_abs(px):
        ptree = px.ptree
        if pf(px):
            px.ptree = ptree
            return True
        return False
    return match_abs


def pSkip():  # @skip()
    def skip(px):
        px.pos = min(px.headpos, px.epos)
        return True
    return skip

# State


State = namedtuple('State', 'sid val prev')


def getstate(state, sid):
    while state is not None:
        if state.sid == sid:
            return state
        state = state.prev
    return None


def pSymbol(pf, sid):  # @symbol(A)
    def match_symbol(px):
        pos = px.pos
        if pf(px):
            px.state = State(sid, px.inputs[pos:px.pos], px.state)
            return True
        return False
    return match_symbol


def pScope(pf):
    def scope(px):
        state = px.state
        res = pf(px)
        px.state = state
        return res
    return scope


def pExists(sid):  # @Match(A)
    return lambda px: getstate(px.state, sid) != None


def pMatch(sid):  # @Match(A)
    def match(px):
        state = getstate(px.state, sid)
        if state is not None and px.inputs.startswith(state.val, px.pos):
            px.pos += len(state.val)
            return True
        return False
    return match

# params = pe.params
# name = str(params[0])
# pf = self.emit(pe.e, step)


def pDef(name, pf):
    def define_dic(px):
        pos = px.pos
        if pf(px):
            s = px.inputs[pos:px.pos]
            if len(s) == 0:
                return True
            if name not in px.dic:
                px.dic[name] = []
            ss = px.dic[name]
            ss.append(s)
            px.dic[name] = sorted(ss, key=lambda x: len(x))[::-1]
            # print(px.dic[name])
            return True
        return False
    return define_dic


# params = pe.params
# name = str(params[0])

def pIn(name):  # @in(NAME)
    def match_dic(px):
        # print('@matching', name, px.inputs, px.pos)
        if name in px.dic:
            ss = px.dic[name]
            for s in ss:
                if px.inputs.startswith(s, px.pos):
                    # print('@matched', s)
                    px.pos += len(s)
                    return True
            # print('@', px.inputs[px.pos:], ss)
        return False
    return match_dic


# Optimized


def pAndChar(text):
    def match_andchar(px):
        return px.inputs.startswith(text, px.pos)
    return match_andchar


def pNotChar(text):
    def match_notchar(px):
        return not px.inputs.startswith(text, px.pos)
    return match_notchar


def pManyChar(text):
    clen = len(text)

    def match_manychar(px):
        while px.inputs.startswith(text, px.pos):
            px.pos += clen
        return True
    return match_manychar


def pOneManyChar(text):
    clen = len(text)

    def match_OneManychar(px):
        if px.inputs.startswith(text, px.pos):
            px.pos += clen
            while px.inputs.startswith(text, px.pos):
                px.pos += clen
            return True
        return False
    return match_OneManychar


def pOptionChar(text):
    clen = len(text)

    def match_optionchar(px):
        if px.inputs.startswith(text, px.pos):
            px.pos += clen
        return True
    return match_optionchar


def pNotRange(chars, ranges):
    pf = pAndRange(chars, ranges)

    def match_ranges(px):
        return not pf(px)
    return match_ranges


def pManyRange(chars, ranges):
    pf = pAndRange(chars, ranges)

    def match_manybitset(px):
        while pf(px):
            px.pos += 1
        return True
    return match_manybitset


def pOneManyRange(chars, ranges):
    pf = pAndRange(chars, ranges)

    def match_onemanybitset(px):
        c = 0
        while pf(px):
            px.pos += 1
            c += 1
        return c > 0
    return match_onemanybitset


def pOptionRange(chars, ranges):
    pf = pAndRange(chars, ranges)

    def match_optionbitset(px):
        if pf(px):
            px.pos += 1
        return True
    return match_optionbitset

# generate


# PContext


class PContext:
    __slots__ = ['inputs', 'pos', 'epos',
                 'headpos', 'ptree', 'state', 'memo', 'dic']

    def __init__(self, inputs, spos, epos):
        self.inputs = inputs
        self.pos = spos
        self.epos = epos
        self.headpos = spos
        self.ptree = None
        self.state = None
        self.memo = [PMemo() for x in range(1789)]
        self.dic = {}

# ParseTree


def rowcol(urn, inputs, spos):
    inputs = inputs[:spos + (1 if len(inputs) > spos else 0)]
    rows = inputs.split(b'\n' if isinstance(inputs, bytes) else '\n')
    return urn, spos, len(rows), len(rows[-1])-1


def nop(s): return s


UNKNOWN_SOURCE = '(unknown source)'


class ParseTree(list):
    def __init__(self, tag, inputs, spos=0, epos=None, urn=UNKNOWN_SOURCE):
        self.tag_ = tag
        self.inputs_ = inputs
        self.spos_ = spos
        self.epos_ = epos if epos is not None else len(inputs)
        self.urn_ = urn

    def getTag(self):
        return self.tag_

    def __eq__(self, tag):
        return self.tag_ == tag

    def getPosition(self):
        return rowcol(self.urn_, self.inputs_, self.spos_)

    def getEndPosition(self):
        return rowcol(self.urn_, self.inputs_, self.epos_)

    def decode(self):
        inputs, spos, epos = self.inputs_, self.spos_, self.epos_
        LF = b'\n' if isinstance(inputs, bytes) else '\n'
        rows = inputs[:spos + (1 if len(inputs) > spos else 0)]
        rows = rows.split(LF)
        linenum, column = len(rows), len(rows[-1])-1
        begin = inputs.rfind(LF, 0, spos) + 1
        # print('@', spos, begin, inputs)
        end = inputs.find(LF, spos)
        # print('@', spos, begin, inputs)
        if end == -1:
            end = len(inputs)
        # print('@[', begin, spos, end, ']', epos)
        line = inputs[begin:end]  # .replace('\t', '   ')
        mark = []
        endcolumn = column + (epos - spos)
        for i, c in enumerate(line):
            if column <= i and i <= endcolumn:
                mark.append('^' if ord(c) < 256 else '^^')
            else:
                mark.append(' ' if ord(c) < 256 else '  ')
        mark = ''.join(mark)
        return (self.urn_, spos, linenum, column, line, mark)

    def message(self, msg='Syntax Error'):
        urn, pos, linenum, cols, line, mark = self.decode()
        return '{} ({}:{}:{}+{})\n{}\n{}'.format(msg, urn, linenum, cols, pos, line, mark)

    def __eq__(self, tag):
        return self.tag_ == tag

    def isSyntaxError(self):
        return self.tag_ == 'err'

    def keys(self):
        ks = []
        for key in self.__dict__:
            v = self.__dict__[key]
            if isinstance(v, ParseTree):
                ks.append(v)
        return ks

    def subs(self):
        es = []
        for i, child in enumerate(self):
            es.append((child.spos_, '', child))
        for key in self.__dict__:
            v = self.__dict__[key]
            if isinstance(v, ParseTree):
                es.append((v.spos_, key, v))
        es.sort()
        return [(x[1], x[2]) for x in es]

    def isEmpty(self):
        return self.tag_ == 'empty'

    def newEmpty(self):
        return ParseTree('empty', self.inputs_, self.epos_, self.epos_, self.urn_)

    def getNodeSize(self):
        return len(self)

    def getSubNodes(self):
        return list(self)

    def has(self, key):
        if isinstance(key, str):
            return hasattr(self, key) and isinstance(getattr(self, key), ParseTree)
        if isinstance(key, int):
            return key < self.getNodeSize()
        return False

    def get(self, key):
        if not self.has(key):
            return self.newEmpty()
        if isinstance(key, str):
            return getattr(self, key)
        return self[key]

    def set(self, key, t):
        assert isinstance(t, ParseTree)
        if key == '':
            self.append(t)
        else:
            setattr(self, key, t)

    def getToken(self, key=None, default_token=''):
        if key is None:
            s = self.inputs_[self.spos_:self.epos_]
            return s.decode('utf-8') if isinstance(s, bytes) else s
        return self.get(key).getToken() if self.has(key) else default_token

    def substring(self, start=None, end=None):
        if start is None:
            if end is None:
                return self.getToken()
            s = self.inputs_[end.epos_:self.epos_]
        else:
            if end is None:
                s = self.inputs_[self.spos_: start.spos_]
            else:
                s = self.inputs_[start.epos_:end.spos_]
        return s.decode('utf-8') if isinstance(s, bytes) else s

    def __str__(self):
        s = self.inputs_[self.spos_:self.epos_]
        return s.decode('utf-8') if isinstance(s, bytes) else s

    def __repr__(self):
        if self.isSyntaxError():
            return self.message('Syntax Error')
        sb = []
        self.strOut(sb, indent='', tab='')
        return "".join(sb)

    def dump(self, indent='\n', tab='  ', tag=nop, edge=nop, token=nop):
        if self.isSyntaxError():
            print(self.message('Syntax Error'))
        else:
            sb = []
            self.strOut(sb, indent, tab, '', tag, edge, token)
            print("".join(sb))

    def strOut(self, sb, indent='\n  ', tab='  ', prefix='', tag=nop, edge=nop, token=nop):
        sb.append(indent + prefix + "[" + tag(f'#{self.getTag()} '))
        subs = self.subs()
        if len(subs) > 0:
            next_indent = indent + tab
            for label, child in subs:
                prefix = edge(label) + ': ' if label != '' else ''
                child.strOut(sb, next_indent, tab, prefix, tag, edge, token)
            sb.append(indent + "]")
        else:
            sb.append(token(repr(str(self))))
            sb.append("]")


def PTreeConv(pt: PTree, urn, inputs):
    if pt.prev != None:
        ct = pt
        while ct.prev is not None:
            ct = ct.prev
        return PTreeConvNode('', urn, inputs, ct.spos, pt.epos, pt)
    if pt.isEdge():
        ct = PTreeConv(pt.child, urn, inputs)
        t = ParseTree('', inputs, ct.spos_, ct.epos_, urn)
        t.set(pt.tag, ct)
        return t
    else:
        return PTreeConvNode(pt.tag, urn, inputs, pt.spos, pt.epos, pt.child)


def PTreeConvNode(tag, urn, inputs, spos, epos, subnode):
    t = ParseTree(tag, inputs, spos, epos, urn)
    while subnode != None:
        if subnode.isEdge():
            # if subnode.child == None:
            #     tt = PTreeConvNode(
            #         '', urn, inputs, subnode.spos, abs(subnode.epos), None)
            # else:
            assert subnode.child != None
            tt = PTreeConv(subnode.child, urn, inputs)
            t.set(subnode.tag, tt)
        else:
            t.append(PTreeConvNode(subnode.tag, urn, inputs,
                                   subnode.spos, abs(subnode.epos), subnode.child))
        subnode = subnode.prev
    for i in range(len(t)//2):
        t[i], t[-(1+i)] = t[-(1+i)], t[i]
    return t


def generate(pf):
    # pf = self.generated[start.uname()]
    def parse(inputs, urn='(unknown source)', pos=0, epos=None, conv=PTreeConv):
        if epos is None:
            epos = len(inputs)
        px = PContext(inputs, pos, epos)
        if not pf(px):
            result = PTree(None, "err", px.headpos, px.headpos, None)
        else:
            result = px.ptree if px.ptree is not None else PTree(None,
                                                                 "", pos, px.pos, None)
        return conv(result, urn, inputs)
    return parse



from pegtree.pasm import ParseTree

try:
  from graphviz import Digraph
  is_graphviz_avaiable = True
except ModuleNotFoundError:
  is_graphviz_avaiable = False


class GenId(object):
  def __init__(self):
    self.c = 0

  def gen(self):
    self.c += 1
    return self.c


def draw_node(g, t, c, fontsize):
  nid = c.gen()
  nodeid = f'n{nid}'
  g.node(nodeid, label=f'#{t.getTag()}', fontsize=fontsize)
  es = []
  for i, child in enumerate(t):
    es.append((child.spos_, child, i))
  for key in t.keys():
    child = t.get(key)
    es.append((child.spos_, child, key))
  es.sort()
  if len(es) == 0:
    childid = f'c{nid}'
    g.node(childid, label=repr(t.getToken()), shape='rectangle', fontsize=fontsize)
    g.edge(nodeid, childid)
  else:
    for i, e in enumerate(es):
      _, child, edge = e
      childid = draw_node(g, child, c, fontsize)
      g.edge(nodeid, childid, label=f'{edge}', fontsize=fontsize)
  return nodeid


def draw_graph(ptree: ParseTree, name='G', fontsize=12):
  if is_graphviz_avaiable == False:
    print('Install graphviz FIRST')
    return
  g = Digraph(name, format='png')
  g.attr('node', fontname='MS Gothic', fontsize=fontsize)
  g.attr('edge', fontname='MS Gothic', fontsize=fontsize)
  draw_node(g, ptree, GenId(), fontsize)
  #g.view()
  return g

def Viz(tree_or_parser, fontsize=12):
  if isinstance(tree_or_parser, ParseTree):
    return draw_graph(tree_or_parser, fontsize)
  return lambda s: draw_graph(tree_or_parser(s), fontsize)



import unittest
import pegtree.cj as cj


S='''
青白く丸い人が寄り道する
大きなノッポの古時計をみた
ハワイについて述べました
ハワイに着いて考えられる
'''.split('\n')

ADJ='''高い
高くない
高過ぎる
高め
高める
高さ
'''.split('\n')

def t(s): return cj.tokenize(s)[0].data()

class TestCJ(unittest.TestCase):

    def test_parse(self):
      for s in S:
        print(repr(cj.parse(s)))

    def test_tokenize(self):
      for s in S:
        print(cj.tokenize(s))

    def test_adj(self):
      self.assertTupleEqual(t('高い'), ('高い', 'A'))
      self.assertTupleEqual(t('高くない'), ('高い', 'A', '@not'))
      self.assertTupleEqual(t('高かった'), ('高い', 'A', '@past'))
      self.assertTupleEqual(t('高くなかった'), ('高い', 'A', '@not', '@past'))
      self.assertTupleEqual(t('高め'), ('高め', 'N'))
      self.assertTupleEqual(t('高める'), ('高める', 'V1'))
      self.assertTupleEqual(t('高さ'), ('高さ', 'N'))
      self.assertTupleEqual(t('青白い'), ('青白い', 'A'))
      self.assertTupleEqual(t('美味しい'), ('美味しい', 'A'))
      self.assertTupleEqual(t('おいしい'), ('おいしい', 'A'))
      print(t('情けない'))
    def test_verb(self):
      print(cj.verb('勝つ', 'polite'))
      print(cj.verb('勝つ', 'not'))
      print(cj.verb('勝つ', 'past'))



if __name__ == '__main__':
    unittest.main()



import unittest
#from test_pegtree import TestTPEG
from test.test_pegtree import TestPEGTree
from test.test_pasm import TestPAsm
from test.test_main import TestMain
from test.test_peg import TestPeg


def suite():
    suite = unittest.TestSuite()
    suite.addTests(unittest.makeSuite(TestPAsm))
    suite.addTests(unittest.makeSuite(TestPeg))
    suite.addTests(unittest.makeSuite(TestPEGTree))
    suite.addTests(unittest.makeSuite(TestMain))
    return suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    test_suite = suite()
    runner.run(test_suite)



import unittest
#import tests
from pegtree import grammar, generate, ParseTree


class TestPEGTree(unittest.TestCase):

    def test_int(self):
        peg = grammar("math.tpeg")
        parser = generate(peg)
        tree = parser('1+')
        self.assertEqual(str(tree), '1')
        self.assertEqual(repr(tree), "[#Int '1']")

    def test_math(self):
        peg = grammar("math.tpeg")
        parser = generate(peg)
        tree = parser('1+2*3')
        self.assertEqual(str(tree), '1+2*3')
        self.assertEqual(
            repr(tree), "[#Add [#Int '1'][#Mul [#Int '2'][#Int '3']]]")
        tree = parser('1*2+3')
        self.assertEqual(str(tree), '1*2+3')
        self.assertEqual(
            repr(tree), "[#Add [#Mul [#Int '1'][#Int '2']][#Int '3']]")


if __name__ == '__main__':
    unittest.main()



import unittest
from pegtree.pasm import *


class TestPAsm(unittest.TestCase):

    def test_Empty(self):
        parser = generate(pChar(''))
        tree = parser('aa')
        self.assertEqual(str(tree), '')
        self.assertEqual(repr(tree), "[# '']")

    def test_Char(self):
        parser = generate(pChar('a'))
        tree = parser('aa')
        self.assertEqual(str(tree), 'a')
        self.assertEqual(repr(tree), "[# 'a']")

    def test_Node(self):
        parser = generate(pNode(pChar('a'), 'A', 0))
        tree = parser('aa')
        self.assertEqual(str(tree), 'a')
        self.assertEqual(repr(tree), "[#A 'a']")

    def test_NullFold(self):
        parser = generate(pFold('', pChar('a'), 'A', 0))
        tree = parser('aa')
        self.assertEqual(str(tree), 'a')
        self.assertEqual(repr(tree), "[#A 'a']")

    def test_NestedNode(self):
        parser = generate(pNode(pSeq2(pNode(pChar('a'), 'A', 0),
                                      pNode(pChar('b'), 'B', 0)), "AB", 0))
        tree = parser('ab')
        self.assertEqual(str(tree), 'ab')
        self.assertEqual(repr(tree), "[#AB [#A 'a'][#B 'b']]")

    def test_FoldNode(self):
        parser = generate(pSeq2(pNode(pChar('a'), 'A', 0),
                                pFold('', pChar('b'), 'B', 0)))
        tree = parser('ab')
        self.assertEqual(str(tree), 'ab')
        self.assertEqual(repr(tree), "[#B [#A 'a']]")

    def test_NodeNode(self):
        parser = generate(pSeq2(pNode(pChar('a'), 'A', 0),
                                pNode(pChar('b'), 'B', 0)))
        tree = parser('ab')
        #print('@', repr(tree))
        self.assertEqual(str(tree), 'ab')
        self.assertEqual(repr(tree), "[# [#A 'a'][#B 'b']]")
        # print(tree.dump())

    def test_NodeNodeFold(self):
        parser = generate(pSeq3(pNode(pChar('a'), 'A', 0),
                                pNode(pChar('b'), 'B', 0),
                                pFold('', pChar('c'), 'C', 0)))
        tree = parser('abc')
        #print('@', repr(tree))
        self.assertEqual(str(tree), 'abc')
        self.assertEqual(repr(tree), "[# [#A 'a'][#C [#B 'b']]]")
        # print(tree.dump())

    def test_NodeFoldFold(self):
        parser = generate(pSeq3(pNode(pChar('a'), 'A', 0),
                                pFold('', pChar('b'), 'B', 0),
                                pFold('', pChar('c'), 'C', 0)))
        tree = parser('abc')
        #print('@', repr(tree))
        self.assertEqual(str(tree), 'abc')
        self.assertEqual(repr(tree), "[#C [#B [#A 'a']]]")
        # print(tree.dump())

    def test_NodeEdgeNode(self):
        parser = generate(pNode(pEdge('x', pNode(pChar('a'), 'A', 0)), 'B', 0))
        tree = parser('abc')
        #print('@', repr(tree))
        self.assertEqual(str(tree), 'a')
        self.assertEqual(repr(tree), "[#B x: [#A 'a']]")

    def test_NodeEdge(self):
        parser = generate(pNode(pEdge('x', pChar('a')), 'A', 0))
        tree = parser('abc')
        #print('@', repr(tree))
        self.assertEqual(str(tree), 'a')
        self.assertEqual(repr(tree), "[#A x: [# 'a']]")

    def test_Edge(self):
        parser = generate(pEdge('x', pChar('a')))
        tree = parser('abc')
        #print('@', repr(tree))
        self.assertEqual(str(tree), 'a')
        self.assertEqual(repr(tree), "[# x: [# 'a']]")

    def test_EdgeEdge(self):
        parser = generate(pEdge('x', pEdge('y', pChar('a'))))
        tree = parser('abc')
        #print('@', repr(tree))
        self.assertEqual(str(tree), 'a')
        self.assertEqual(repr(tree), "[# x: [# y: [# 'a']]]")


if __name__ == '__main__':
    unittest.main()



import sys
import os
import errno
import inspect
from pathlib import Path
from .peg import *
from .pasm import generate
from .tpeg_pasm import TPEGGrammar
#from .terminal import DefaultConsole as console

from .visitor import Visitor

#
# BuiltIn_NonTerminal
# 

BuiltIn_NonTerminal = {
    'W': PRange('_', 'AZaz09'),
    'LF': POre(PChar('\r\n'), PChar('\n')),
    'EOF': PNot(PAny()),
    '_': PMany(PRange(' \t', '')),
    '__': PMany(PRange(' \t\r\n', '')),
}

TPEGParser = generate(TPEGGrammar['Start'])


class TPEGLoader(Visitor):
    def __init__(self, peg, **options):
        self.names = {}
        self.peg = peg
        self.options = options
        self.isOnlyPEG = options.get('isPurePEG', False)

    def load(self, ptree):
        for stmt in ptree:
            if stmt == 'Rule':
                name = str(stmt.name)
                if name in self.names:
                    # pos4 = stmt['name'].getpos4()
                    self.warning(stmt.name, f'redefined name {name}')
                    continue
                self.names[name] = stmt.e
            elif stmt == 'Example':
                doc = stmt.doc
                for n in stmt.names:
                    self.example(str(n), doc)
        for name in self.names:
            ptree = self.names[name]
            self.peg[name] = self.visit(ptree)

    def example(self, name, doc):
        self.peg['@@example'].append((name, doc))

    # def conv(self, ptree):
    #     tag = ptree.getTag()
    #     pe = ptree
    #     if hasattr(self, tag):
    #         f = getattr(self, tag)
    #         pe = f(ptree)
    #     if not isinstance(pe, PExpr):
    #         print('FIXME(TPEGLoader(conv)', tag, type(pe), pe)
    # return pe

    def acceptEmpty(self, ptree):
        return EMPTY

    def acceptAny(self, ptree):
        return ANY

    def acceptChar(self, ptree):
        s = ptree.getToken()
        sb = []
        while len(s) > 0:
            c, s = TPEGLoader.unquote(s)
            sb.append(c)
        return PChar(''.join(sb))

    @classmethod
    def unquote(cls, s):
        if s.startswith('\\'):
            if s.startswith('\\n'):
                return '\n', s[2:]
            if s.startswith('\\t'):
                return '\t', s[2:]
            if s.startswith('\\r'):
                return '\r', s[2:]
            if s.startswith('\\v'):
                return '\v', s[2:]
            if s.startswith('\\f'):
                return '\f', s[2:]
            if s.startswith('\\b'):
                return '\b', s[2:]
            if (s.startswith('\\x') or s.startswith('\\X')) and len(s) >= 4:
                c = int(s[2:4], 16)
                return chr(c), s[4:]
            if (s.startswith('\\u') or s.startswith('\\U')) and len(s) >= 6:
                c = chr(int(s[2:6], 16))
                if len(c) != 1:
                    c = ''  # remove unsupported unicode
                return c, s[6:]
            else:
                return s[1], s[2:]
        else:
            return s[0], s[1:]

    def acceptClass(self, ptree):
        s = ptree.getToken()
        chars = []
        ranges = []
        while len(s) > 0:
            c, s = TPEGLoader.unquote(s)
            if s.startswith('-') and len(s) > 1:
                c2, s = TPEGLoader.unquote(s[1:])
                ranges.append(c + c2)
            else:
                chars.append(c)
        if len(chars) == 0 and len(ranges) == 0:
            return FAIL
        if len(chars) == 1 and len(ranges) == 0:
            return PChar(chars[0])
        return PRange(''.join(chars), ''.join(ranges))

    def newRef(self, name):
        if name in self.names:
            return self.peg.newRef(name)
        if name in BuiltIn_NonTerminal:
            return BuiltIn_NonTerminal[name]
        es = [PChar(name)]
        if len(name) > 0 and name[-1].isalnum():
            es.append(PNot(self.newRef('W')))
        es.append(self.newRef('_'))
        return PSeq(*es)

    def acceptName(self, ptree):
        name = ptree.getToken()
        if name in self.names:
            ref = self.peg.newRef(name)
            return PName(ref, ref.uname(), ptree)
        if name in BuiltIn_NonTerminal:
            return self.newRef(name)
        if name[0].isupper() or name.startswith('_'):  # or name[0].islower() :
            self.perror(ptree, f'undefined nonterminal {name}')
            self.peg['@error'] = True
            self.peg[name] = EMPTY
            return self.peg.newRef(name)
        self.warning(ptree, f'undefined nonterminal {name}')
        return self.newRef(name)

    def acceptQuoted(self, ptree):
        name = ptree.getToken()
        if name in self.names:
            ref = self.peg.newRef(name)
            return PName(ref, ref.uname(), ptree)
        self.warning(ptree, f'undefined terminal {name}')
        name = name[1:-1]
        return self.newRef(name)

    def acceptMany(self, ptree):
        return PMany(self.visit(ptree.e))

    # def acceptMany1(self, ptree):
    #     return POneMany(self.visit(ptree.e))

    def acceptOneMany(self, ptree):
        return POneMany(self.visit(ptree.e))

    def acceptOption(self, ptree):
        return POption(self.visit(ptree.e))

    def acceptAnd(self, ptree):
        return PAnd(self.visit(ptree.e))

    def acceptNot(self, ptree):
        return PNot(self.visit(ptree.e))

    def acceptSeq(self, ptree):
        return PSeq.new(*tuple(map(lambda p: self.visit(p), ptree)))

    def acceptOre(self, ptree):
        return POre.new(*tuple(map(lambda p: self.visit(p), ptree)))

    def acceptAlt(self, ptree):
        self.warning(ptree, f'unordered choice is not supported')
        return POre.new(*tuple(map(lambda p: self.visit(p), ptree)))

    def acceptNode(self, ptree):
        tag = ptree.getToken('tag', '')
        e = self.visit(ptree.e)
        return e if self.isOnlyPEG else PNode(e, tag, 0) 

    def acceptEdge(self, ptree):
        edge = ptree.getToken('edge', '')
        e = self.visit(ptree.e)
        return e if self.isOnlyPEG else PEdge(edge, e)

    def acceptFold(self, ptree):
        edge = ptree.getToken('edge', '')
        tag = ptree.getToken('tag', '')
        e = self.visit(ptree.e)
        return e if self.isOnlyPEG else PFold(edge, e, tag, 0)

    FIRST = {'lazy', 'scope', 'symbol', 'def',
             'match', 'equals', 'contains', 'cat'}

    def acceptFunc(self, ptree):
        funcname = ptree.getToken(0)
        ps = [self.visit(p) for p in ptree[1:]]
        # if funcname.startswith('import'):
        #     return TPEGLoader.loadImport(ptree, ps)
        # if funcname.startswith('choice'):
        #     return TPEGLoader.loadChoice(ptree, funcname, ps)
        if funcname in TPEGLoader.FIRST:
            return ps[0] if self.isOnlyPEG else PAction(ps[0], funcname, tuple(ps), ptree)
        return EMPTY if self.isOnlyPEG else PAction(EMPTY, funcname, tuple(ps), ptree)

    # @classmethod
    # def loadImport(cls, ptree, ps):
    #     urn = str(ps[0])
    #     apeg = grammar(urn)
    #     if len(ps) > 1:
    #         name = str(ps[1])
    #         if not name in apeg:
    #             logger()
    #             return EMPTY
    #         return apeg.newRef(name)
    #     else:
    #         apeg.start()

    #     return TPEGLoader.choice(ptree.urn_, ps, 0, fgt)

    # @classmethod
    # def loadChoice(cls, ptree, funcname, ps):
    #     def feq(ss, n):
    #         return {x for x in ss if len(x) == n and not x.startswith('#')}

    #     def fgt(ss, n):
    #         return {x for x in ss if len(x) > n and not x.startswith('#')}
    #     n = funcname[6:]
    #     if n.isdigit():
    #         return TPEGLoader.choice(ptree.urn_, ps, int(n), feq)
    #     if n.startswith('G'):
    #         return TPEGLoader.choice(ptree.urn_, ps, int(n[1:]), fgt)
    #     return TPEGLoader.choice(ptree.urn_, ps, 0, fgt)

    # @classmethod
    # def fileName(cls, urn, file):
    #     if file.startswith('CJDIC'):
    #         file = file.replace('CJDIC', os.environ.get('CJDIC', '__unknown__'))
    #         return None if '__unknown__' in file else Path(file)
    #     return Path(urn).parent / file

    # @classmethod
    # def choice(cls, urn, es, n, fset):
    #     ds = set()
    #     for e in es:
    #         filename = str(e)[1:-1]
    #         file = TPEGLoader.fileName(urn, filename)
    #         if file is None:
    #             continue
    #         try:
    #             with file.open(encoding='utf-8_sig') as f:
    #                 ss = [x.strip('\r\n') for x in f.readlines()]
    #                 ds |= fset(ss, n)
    #         except:
    #             print(f'can\'t read: {filename} {file}')
    #     choice = [PChar(x) for x in sorted(ds, key=lambda x: len(x))[::-1]]
    #     return POre.new(*choice)

# class LeftRef(PVisitor):
#     EMPTYSET = set()

#     def __init__(self, checkLeftRec=True):
#         self.checkLeftRec = checkLeftRec
#         self.memos = {}

#     def PChar(self, pe): return LeftRef.EMPTYSET
#     def PAny(self, pe): return LeftRef.EMPTYSET
#     def PRange(self, pe): return LeftRef.EMPTYSET

#     def PRef(self, pe):
#         uname = pe.uname()
#         if uname not in self.memos:
#             memos[uname] = LeftRef.EMPTYSET
#             memos[uname] = self.visit(pe.deref())
#         return set(pe)

#     def PName(self, pe):
#         if self.checkLeftRec:
#             self.visit(pe.e)
#             return set(pe)
#         return self.visit(pe.e)

#     def PAnd(self, pe): return self.visit(pe.e)
#     def PNot(self, pe): return self.visit(pe.e)
#     def PMany(self, pe): return self.visit(pe.e)
#     def POneMany(self, pe): return self.visit(pe.e)
#     def POption(self, pe): return self.visit(pe.e)

#     def PSeq(self, pe):
#         result = set()
#         for e in pe:
#             result |= self.visit(e)
#             if self.checkLeftRec & isAlwaysConsumed(e):
#                 break
#         return result

#     def POre(self, pe):
#         result = set()
#         for e in pe:
#             result |= self.visit(e)
#         return result

#     def PNode(self, pe): return self.visit(pe.e)
#     def PFold(self, pe): return self.visit(pe.e)
#     def PEdge(self, pe): return self.visit(pe.e)
#     def PAbs(self, pe): return self.visit(pe.e)

#     def PAction(self, pe): return self.visit(pe.e)

#def checkLeftRecursion(peg: Grammar):


def load_grammar(peg, file_or_text, **options):
    # pegparser = pasm.generate(options.get('peg', TPEGGrammar))
    
    if isinstance(file_or_text, Path) and file_or_text.is_file():
        f = file_or_text.open(encoding=options.get('encoding', 'utf-8_sig'))
        data = f.read()
        f.close()
        basepath = str(file_or_text)
        ptree = TPEGParser(data, basepath)
    else:
        if 'basepath' in options:
            basepath = options['basepath']
        else:
            basepath = inspect.currentframe().f_back.f_code.co_filename
        ptree = TPEGParser(file_or_text, options.get('urn', basepath))
        basepath = (str(Path(basepath).resolve().parent))
    options['basepath'] = basepath
    if ptree.isSyntaxError():
        console.perror(ptree, 'Syntax Error')
        peg['@error'] = True
        return
    pconv = TPEGLoader(peg, **options)
    pconv.load(ptree)

def findpath(paths, file_or_text):
    if file_or_text.find('=') > 0 or file_or_text.find('<-') > 0:
        return file_or_text
    for p in paths:
        path = Path(p) / file_or_text
        if path.is_file():
            stat = path.stat()
            return path.absolute(), stat.st_mtime
    raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), file_or_text)

GrammarDB = {}

def grammar(file_or_text, **options):
    paths = []
    basepath = options.get('basepath', '')
    if basepath == '':
        paths.append('')
    else:
        paths.append(str(Path(basepath).resolve().parent))
    framepath = inspect.currentframe().f_back.f_code.co_filename
    paths.append(str(Path(framepath).resolve().parent))
    paths.append(str(Path(__file__).resolve().parent / 'grammar'))
    paths += os.environ.get('GRAMMAR', '').split(':')
    path_or_text, mtime = findpath(paths, file_or_text)
    key = f'{path_or_text}:{mtime}'
    if key in GrammarDB:
        return GrammarDB[key]
    peg = Grammar()
    load_grammar(peg, path_or_text, **options)
    GrammarDB[key] = peg
    return peg

if __name__ == '__main__':
    peg = grammar('es4.tpeg')



import sys
import os
import errno
import inspect
from pathlib import Path
from pegtree.peg import *
import pegtree.pasm as pasm
from pegtree.tpeg_pasm import TPEGGrammar
from pegtree.terminal import DefaultConsole as console

#
# BuiltIn_NonTerminal
# 

BuiltIn_NonTerminal = {
    'W': PRange('_', 'AZaz09'),
    'LF': POre(PChar('\r\n'), PChar('\n')),
    'EOF': PNot(PAny()),
    '_': PMany(PRange(' \t', '')),
    '__': PMany(PRange(' \t\r\n', '')),
}


class TPEGLoader(object):
    def __init__(self, peg, **options):
        self.names = {}
        self.peg = peg
        self.isOnlyPEG = options.get('isPurePEG', False)

    def load(self, ptree):
        for stmt in ptree:
            if stmt == 'Rule':
                name = str(stmt.name)
                if name in self.names:
                    # pos4 = stmt['name'].getpos4()
                    console.perror(stmt.name, f'redefined name {name}')
                    continue
                self.names[name] = stmt.e
            elif stmt == 'Example':
                doc = stmt.doc
                for n in stmt.names:
                    self.example(str(n), doc)
            # elif stmt == 'Import':
            #     urn = str(stmt.name)
            #     apeg = grammar(urn, **options)
            #     for n in stmt.names:
            #         lname = str(n)  # ns.Expression
            #         name = lname
            #         if lname.find('.') != -1:
            #             name = lname.split('.')[-1]
            #         pos4 = n.getpos4()
            #         if not name in apeg:
            #             logger('perror', pos4, f'undefined name {name}')
            #             continue
            #         g.add(lname, Action(apeg.newRef(name),
            #                             'import', (name, urn), pos4))
        for name in self.names:
            ptree = self.names[name]
            self.peg[name] = self.conv(ptree)

    def example(self, name, doc):
        self.peg['@@example'].append((name, doc))

    def conv(self, ptree):
        tag = ptree.getTag()
        pe = ptree
        if hasattr(self, tag):
            f = getattr(self, tag)
            pe = f(ptree)
        if not isinstance(pe, PExpr):
            print('FIXME(TPEGLoader(conv)', tag, type(pe), pe)
        return pe

    def Empty(self, ptree):
        return EMPTY

    def Any(self, ptree):
        return ANY

    def Char(self, ptree):
        s = ptree.getToken()
        sb = []
        while len(s) > 0:
            c, s = TPEGLoader.unquote(s)
            sb.append(c)
        return PChar(''.join(sb))

    @classmethod
    def unquote(cls, s):
        if s.startswith('\\'):
            if s.startswith('\\n'):
                return '\n', s[2:]
            if s.startswith('\\t'):
                return '\t', s[2:]
            if s.startswith('\\r'):
                return '\r', s[2:]
            if s.startswith('\\v'):
                return '\v', s[2:]
            if s.startswith('\\f'):
                return '\f', s[2:]
            if s.startswith('\\b'):
                return '\b', s[2:]
            if (s.startswith('\\x') or s.startswith('\\X')) and len(s) >= 4:
                c = int(s[2:4], 16)
                return chr(c), s[4:]
            if (s.startswith('\\u') or s.startswith('\\U')) and len(s) >= 6:
                c = chr(int(s[2:6], 16))
                if len(c) != 1:
                    c = ''  # remove unsupported unicode
                return c, s[6:]
            else:
                return s[1], s[2:]
        else:
            return s[0], s[1:]

    def Class(self, ptree):
        s = ptree.getToken()
        chars = []
        ranges = []
        while len(s) > 0:
            c, s = TPEGLoader.unquote(s)
            if s.startswith('-') and len(s) > 1:
                c2, s = TPEGLoader.unquote(s[1:])
                ranges.append(c + c2)
            else:
                chars.append(c)
        if len(chars) == 0 and len(ranges) == 0:
            return FAIL
        if len(chars) == 1 and len(ranges) == 0:
            return PChar(chars[0])
        return PRange(''.join(chars), ''.join(ranges))

    def newRef(self, name):
        if name in self.names:
            return self.peg.newRef(name)
        if name in BuiltIn_NonTerminal:
            return BuiltIn_NonTerminal[name]
        es = [PChar(name)]
        if len(name) > 0 and name[-1].isalnum():
            es.append(PNot(self.newRef('W')))
        es.append(self.newRef('_'))
        return PSeq(*es)

    def Name(self, ptree):
        name = ptree.getToken()
        if name in self.names:
            ref = self.peg.newRef(name)
            return PName(ref, ref.uname(), ptree)
        if name in BuiltIn_NonTerminal:
            return self.newRef(name)
        if name[0].isupper() or name.startswith('_'):  # or name[0].islower() :
            console.perror(ptree, f'undefined nonterminal {console.bold(name)}')
            self.peg['@error'] = True
            self.peg[name] = EMPTY
            return self.peg.newRef(name)
        console.pwarn(ptree, f'undefined nonterminal {console.bold(name)}')
        return self.newRef(name)

    def Quoted(self, ptree):
        name = ptree.getToken()
        if name in self.names:
            ref = self.peg.newRef(name)
            return PName(ref, ref.uname(), ptree)
        console.pwarn(ptree, f'undefined terminal {console.bold(name)}')
        name = name[1:-1]
        return self.newRef(name)

    def Many(self, ptree):
        return PMany(self.conv(ptree.e))

    def Many1(self, ptree):
        return POneMany(self.conv(ptree.e))

    def OneMany(self, ptree):
        return POneMany(self.conv(ptree.e))

    def Option(self, ptree):
        return POption(self.conv(ptree.e))

    def And(self, ptree):
        return PAnd(self.conv(ptree.e))

    def Not(self, ptree):
        return PNot(self.conv(ptree.e))

    def Seq(self, ptree):
        return PSeq.new(*tuple(map(lambda p: self.conv(p), ptree)))

    def Ore(self, ptree):
        return POre.new(*tuple(map(lambda p: self.conv(p), ptree)))

    def Alt(self, ptree):
        self.pwarn(ptree, f'unordered choice is not supported')
        return POre.new(*tuple(map(lambda p: self.conv(p), ptree)))

    def Node(self, ptree):
        tag = ptree.getToken('tag', '')
        e = self.conv(ptree.e)
        return e if self.isOnlyPEG else PNode(e, tag, 0) 

    def Edge(self, ptree):
        edge = ptree.getToken('edge', '')
        e = self.conv(ptree.e)
        return e if self.isOnlyPEG else PEdge(edge, e)

    def Fold(self, ptree):
        edge = ptree.getToken('edge', '')
        tag = ptree.getToken('tag', '')
        e = self.conv(ptree.e)
        return e if self.isOnlyPEG else PFold(edge, e, tag, 0)

    FIRST = {'lazy', 'scope', 'symbol', 'def',
             'match', 'equals', 'contains', 'cat'}

    def Func(self, ptree):
        funcname = ptree.getToken(0)
        ps = [self.conv(p) for p in ptree[1:]]
        if funcname.startswith('choice'):
          return TPEGLoader.loadChoice(ptree, funcname, ps)
        if funcname in TPEGLoader.FIRST:
            return ps[0] if self.isOnlyPEG else PAction(ps[0], funcname, tuple(ps), ptree)
        return EMPTY if self.isOnlyPEG else PAction(EMPTY, funcname, tuple(ps), ptree)

    @classmethod
    def loadChoice(cls, ptree, funcname, ps):
        def feq(ss, n):
            return {x for x in ss if len(x) == n and not x.startswith('#')}

        def fgt(ss, n):
            return {x for x in ss if len(x) > n and not x.startswith('#')}
        n = funcname[6:]
        if n.isdigit():
            return TPEGLoader.choice(ptree.urn_, ps, int(n), feq)
        if n.startswith('G'):
            return TPEGLoader.choice(ptree.urn_, ps, int(n[1:]), fgt)
        return TPEGLoader.choice(ptree.urn_, ps, 0, fgt)

    @classmethod
    def fileName(cls, urn, file):
        if file.startswith('CJDIC'):
            file = file.replace('CJDIC', os.environ.get('CJDIC', '__unknown__'))
            return None if '__unknown__' in file else Path(file)
        return Path(urn).parent / file

    @classmethod
    def choice(cls, urn, es, n, fset):
        ds = set()
        for e in es:
            filename = str(e)[1:-1]
            file = TPEGLoader.fileName(urn, filename)
            if file is None:
                continue
            try:
                with file.open(encoding='utf-8_sig') as f:
                    ss = [x.strip('\r\n') for x in f.readlines()]
                    ds |= fset(ss, n)
            except:
                print(f'can\'t read: {filename} {file}')
        choice = [PChar(x) for x in sorted(ds, key=lambda x: len(x))[::-1]]
        return POre.new(*choice)

class LeftRef(PVisitor):
    EMPTYSET = set()

    def __init__(self, checkLeftRec=True):
        self.checkLeftRec = checkLeftRec
        self.memos = {}

    def PChar(self, pe): return LeftRef.EMPTYSET
    def PAny(self, pe): return LeftRef.EMPTYSET
    def PRange(self, pe): return LeftRef.EMPTYSET

    def PRef(self, pe):
        uname = pe.uname()
        if uname not in self.memos:
            memos[uname] = LeftRef.EMPTYSET
            memos[uname] = self.visit(pe.deref())
        return set(pe)

    def PName(self, pe):
        if self.checkLeftRec:
            self.visit(pe.e)
            return set(pe)
        return self.visit(pe.e)

    def PAnd(self, pe): return self.visit(pe.e)
    def PNot(self, pe): return self.visit(pe.e)
    def PMany(self, pe): return self.visit(pe.e)
    def POneMany(self, pe): return self.visit(pe.e)
    def POption(self, pe): return self.visit(pe.e)

    def PSeq(self, pe):
        result = set()
        for e in pe:
            result |= self.visit(e)
            if self.checkLeftRec & isAlwaysConsumed(e):
                break
        return result

    def POre(self, pe):
        result = set()
        for e in pe:
            result |= self.visit(e)
        return result

    def PNode(self, pe): return self.visit(pe.e)
    def PFold(self, pe): return self.visit(pe.e)
    def PEdge(self, pe): return self.visit(pe.e)
    def PAbs(self, pe): return self.visit(pe.e)

    def PAction(self, pe): return self.visit(pe.e)

#def checkLeftRecursion(peg: Grammar):


TPEGParser = pasm.generate(TPEGGrammar['Start'])

def load_grammar(peg, file_or_text, **options):
    # pegparser = pasm.generate(options.get('peg', TPEGGrammar))
    
    if isinstance(file_or_text, Path) and file_or_text.is_file():
        f = file_or_text.open(encoding=options.get('encoding', 'utf-8_sig'))
        data = f.read()
        f.close()
        basepath = str(file_or_text)
        ptree = TPEGParser(data, basepath)
    else:
        if 'basepath' in options:
            basepath = options['basepath']
        else:
            basepath = inspect.currentframe().f_back.f_code.co_filename
        ptree = TPEGParser(file_or_text, options.get('urn', basepath))
        basepath = (str(Path(basepath).resolve().parent))
    options['basepath'] = basepath
    if ptree.isSyntaxError():
        console.perror(ptree, 'Syntax Error')
        peg['@error'] = True
        return
    pconv = TPEGLoader(peg, **options)
    pconv.load(ptree)

def findpath(paths, file_or_text):
    if file_or_text.find('=') > 0 or file_or_text.find('<-') > 0:
        return file_or_text
    for p in paths:
        path = Path(p) / file_or_text
        if path.is_file():
            return path.absolute()
    raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), file_or_text)

GrammarDB = {}

def grammar(file_or_text, **options):
    paths = []
    basepath = options.get('basepath', '')
    if basepath == '':
        paths.append('')
    else:
        paths.append(str(Path(basepath).resolve().parent))
    framepath = inspect.currentframe().f_back.f_code.co_filename
    paths.append(str(Path(framepath).resolve().parent))
    paths.append(str(Path(__file__).resolve().parent / 'grammar'))
    paths += os.environ.get('GRAMMAR', '').split(':')
    path = findpath(paths, file_or_text)
    key = str(path)
    if key in GrammarDB:
        return GrammarDB[key]
    peg = Grammar()
    load_grammar(peg, path, **options)
    if key.endswith('peg') or key.endswith('pegtree'):
        GrammarDB[key] = peg
    return peg

if __name__ == '__main__':
    peg = grammar('es4.tpeg')



import string
from pegtree.pasm import ParseTree


def make_graph(t: ParseTree, nodes, edges, c):
    es = []
    for i, child in enumerate(t):
        es.append((child.spos_, child, i))
    for key in t.keys():
        child = t.get(key)
        es.append((child.spos_, child, key))
    es.sort()
    if len(es) > 0:
        parentId = c
        nodes.append((parentId, t.getTag(), None))
        for i, e in enumerate(es):
            _, child, edge = e
            childId = c+1
            c = make_graph(child, nodes, edges, childId)
            edges.append((parentId, childId, edge))
        return c
    if t != '':
        nodes.append((parentId, t.getTag(), None))
        edges.append((parentId, parentId+1, None))
        nodes.append((parentId+1, None, str(t)))
        return parentId+1
    nodes.append((c, None, str(t)))
    return c

# dot


DOT = '''\
digraph sample {
    graph [
        charset = "UTF-8",
        label = "$input_text",
        labelloc = t,
        fontsize = 18,
        dpi = 300,
    ];

    edge [
        dir = none,
        fontname = "MS Gothic",
        fontcolor = "#252525",
        fontsize = 12,
    ];

    node [
        shape = box,
        style = "rounded,filled",
        color = "#3c3c3c",
        fillcolor = "#f5f5f5",
        fontname = "MS Gothic",
        fontsize = 16,
        fontcolor = "#252525",
    ];

    $node_description

    $edge_description
}'''


def escape(s):
    after = ''
    META_LITERAL = ['\\', '"']
    for c in s:
        if c in META_LITERAL:
            after += f'\\{c}'
        else:
            after += c
    return after


'''
def bottom_check(s):
    for c in s:
        if unicodedata.east_asian_width(c) in ['W', 'F', 'H']:
            return ', labelloc = "bottom"'
    return ''


def make_dot(nodes, edges, t: ParseTree, idc: int):
    tid = idc
    for i, child in enumerate(t):
        idc = make_dot(nodes, edges, child, idc)
        edge = f'n{tid} -> n{idc}   [label="[{0}]"]'
        edges.append(edge)
        hasContent = True
    for key in t.keys():
        idc = make_dot(nodes, edges, t.get(key), idc)
        edge = f'n{tid} -> n{idc}   [label="{key}"]'
        edges.append(edge)
        hasContent = True
    if not hasContent:
        node = f'n{tid} [label="#{t.getTag()}"]'
        nodes.append(node)
        idc = tid+1
        node = f'n{idc} [label="#{t.getTag()}"]'
        edge = f'n{tid} -> n{idc}'
        edges.append(edge)
    return idc
      d['node'].append(f'n{nid} [label="#{t.tag}"]')
       if len(t.subs()) == 0:
            leaf = str(t)
            d['node'].append(
                f'n{nid}_0 [label="{escape(leaf)}"{bottom_check(leaf)}]')
            d['edge'].append(f'n{nid} -> n{nid}_0')
        else:
            for i, (fst, snd) in enumerate(t.subs()):
                label=f' [label="{fst}"]' if fst != '' else ''
                d['edge'].append(f'n{nid} -> n{nid}_{i}{label}')
                make_dot(snd, d, f'{nid}_{i}')
'''


def dot_node(nodeId, tag, token):
    if tag is not None:
        label = f'[label="#{tag}"]'
    else:
        label = f'[label={dot_quote(token)}]'
    return f'n{nodeId} {label}'


def dot_edge(parentId, childId, edge):
    if edge == None:
        label = ''
    else:
        if isinstance(edge, int):
            edge = f'[{edge}]'
        label = f'[label={dot_quote(edge)}]'
    return f'n{parentId} -> n{childId}  {label}'


DOTESC = str.maketrans(
    {'\n': '\\n', '\t': '\\t', '\r': '\\r', '\v': '\\v', '\f': '\\f',
     '\\': '\\\\', "'": "\\'"})


def dot_quote(s):
    return '"' + s.translate(DOTESC) + '"'


def dot(t: ParseTree):
    nodes = []
    edges = []
    make_graph(t, nodes, edges, 0)
    context = {
        'input_text': dot_quote(t.inputs),
        'node_description': ';\n    '.join([dot_node(*x) for x in nodes]),
        'edge_description': ';\n    '.join([dot_edge(*x) for x in nodes]),
    }
    return string.Template(DOT).substitute(context)


def synttree(t, fmt='\\synttree{{1}}{}'):
    sb = []
    sb.append(f'[\\Tag{{{t.getTag()}}}')
    subs = t.subs()
    if len(subs) > 0:
        for _, sub in subs:
            #sb.append(' ')
            sb.append(synttree(sub, None))
    else:
        #sb.append(' ')
        sb.append(f'[\\Token{{{str(t)}}}]')
    sb.append(']')
    s = ' '.join(sb)
    return s if fmt is None else fmt.format(s)


def qtree(t, fmt='\\Tree {}'):
    sb = []
    sb.append(f'[.\\Tag{{{t.getTag()}}}')
    subs = t.subs()
    if len(subs) > 0:
        c = -(len(subs)+1)//2
        for label, sub in subs:
            if label != '':
                pos = 'left' if c <= 0 else 'right'
                sb.append(f'\\edge node[midway,{pos}]{{\\Edge{{{label}}}}};')
            sb.append(qtree(sub, None))
            c += 1
    else:
        #sb.append(' ')
        sb.append(f'\\Token{{{str(t)}}}')
    sb.append(']')
    s = ' '.join(sb)
    return s if fmt is None else fmt.format(s)

import pegtree.cj as cj 
def cjdump(tree):
    s = repr(tree)
    tokens = cj.tokenize(tree)
    return s + '\n' + repr(tokens)

def treedump(options, dump):
    ext = options.get('format')
    if ext == 'dot':
        return dot
    if ext == 'qtree':
        return qtree
    if ext == 'synttree' or ext == 'syntree':
        return synttree
    if ext == 'cj':
        return cjdump
    return dump



import unittest
#import tests
from pegtree.main import main


class TestMain(unittest.TestCase):

    def test_sample(self):
        print('pegtree sample')
        main(['', 'sample'])

    def test_example(self):
        print('pegtree sample')
        main(['', 'example', '-g', 'math.tpeg'])

    def test_peg(self):
        print('pegtree peg -g math.tpeg')
        main(['', 'peg', '-g', 'math.tpeg'])

    def test_pasm(self):
        print('pegtree pasm -g math.tpeg')
        main(['', 'pasm', '-g', 'math.tpeg'])


if __name__ == '__main__':
    unittest.main()



from pegtree.pasm import unique_range
#from pegtree.terminal import DEBUG, VERBOSE

import os

DebugFlag = 'DEBUG' in os.environ
VerboseFlag = 'VERBOSE' in os.environ


def DEBUG(*x):
    if DebugFlag:
        print('DEBUG', *x)


def VERBOSE(*x):
    if DebugFlag:
        print(*x)


# Grammar

GrammarId = 0


class Grammar(dict):
    def __init__(self, source=None):
        global GrammarId
        self.ns = str(GrammarId)
        self.N = []
        GrammarId += 1
        super().__setitem__('@@example', [])
        super().__setitem__('@@refs', {})

    def __repr__(self):
        ss = []
        for rule in self.N:
            ss.append(rule)
            ss.append(' = ')
            ss.append(repr(self[rule]))
            ss.append('\n')
        return ''.join(ss)

    def __setitem__(self, key, item):
        if not key in self:
            self.N.append(key)
        super().__setitem__(key, item)

    def newRef(self, name):
        refs = self['@@refs']
        if name not in refs:
            refs[name] = PRef(self, name)
        return refs[name]

    def start(self):
        if len(self.N) == 0:
            self['EMPTY'] = EMPTY
        return self.N[0]


class PExpr(object):
    def __iter__(self): pass
    def __len__(self): return 0

    def cname(self):
        return self.__class__.__name__


class PUnary(PExpr):
    __slot__ = ['e']

    def __init__(self, e):
        self.e = e

    def __iter__(self):
        yield self.e

    def __len__(self):
        return 1

    def minLen(self): return self.e.minLen()

    def grouping(self, e):
        if isinstance(e, POre) or isinstance(e, PSeq) or isinstance(e, PEdge) or isinstance(e, PAlt):
            return '(' + repr(e) + ')'
        return repr(e)


class PTuple(PExpr):
    __slots__ = ['es']

    def __init__(self, *es):
        self.es = list(es)

    def __iter__(self):
        return iter(self.es)

    def __len__(self):
        return len(self.es)

    def minLen(self):
        if not hasattr(self, 'minlen'):
            self.minlen = min(map(lambda e: e.minLen(), self.es))
        return self.minlen

    def grouping(self, e):
        if isinstance(e, POre) or isinstance(e, PAlt):
            return '(' + repr(e) + ')'
        return repr(e)

# Parsing Expression


class PAny(PExpr):
    def __repr__(self): return '.'
    def minLen(self): return 1


class PChar(PExpr):
    __slots__ = ['text']
    ESCTBL = str.maketrans(
        {'\n': '\\n', '\t': '\\t', '\r': '\\r', '\v': '\\v', '\f': '\\f',
         '\\': '\\\\', "'": "\\'"})

    def __init__(self, text):
        self.text = text

    def __repr__(self):
        return "'" + self.text.translate(PChar.ESCTBL) + "'"

    def minLen(self): return len(self.text)

    @classmethod
    def new(cls, text):
        if len(text) == 0:
            return EMPTY
        return PChar(text)


class PRange(PExpr):
    __slots__ = ['chars', 'ranges', 'neg']
    ESCTBL = str.maketrans(
        {'\n': '\\n', '\t': '\\t', '\r': '\\r', '\v': '\\v', '\f': '\\f',
         '\\': '\\\\', ']': '\\]', '-': '\\-'})

    def __init__(self, chars, ranges, neg=False):
        self.chars = chars
        self.ranges = ranges
        self.neg = neg

    def __repr__(self):
        sb = []
        if(self.neg):
            sb.append('(!')
        sb.append('[')
        sb.append(self.chars.translate(PRange.ESCTBL))
        r = self.ranges
        while len(r) > 1:
            sb.append(r[0].translate(PRange.ESCTBL))
            sb.append('-')
            sb.append(r[1].translate(PRange.ESCTBL))
            r = r[2:]
        sb.append(']')
        if(self.neg):
            sb.append('. )')
        return ''.join(sb)

    def minLen(self): return 1

    @classmethod
    def new(cls, chars, ranges, neg=False):
        if len(chars) == 0 and len(ranges) == 0:
            return ANY if neg else FAIL
        if len(chars) == 1 and len(ranges) == 0 and neg == False:
            return PChar(chars)
        return PRange(chars, ranges, neg)


class PRef(PExpr):
    def __init__(self, peg, name):
        self.peg = peg
        self.name = name

    def __repr__(self):
        return self.name

    def uname(self, peg=None):
        if self.peg == peg:
            return self.name
        return f'{self.peg.ns}{self.name}'

    def deref(self):
        return self.peg[self.name]

    def minLen(self):
        if not hasattr(self, 'minlen'):
            self.minlen = 0
            self.minlen = self.deref().minLen()
        return self.minlen


class PName(PUnary):
    __slot__ = ['e', 'name', 'ptree', 'isLeftRec']

    def __init__(self, e, name: str, ptree, isLeftRec=False):
        super().__init__(e)
        self.name = name
        self.ptree = ptree
        self.isLeftRec = isLeftRec

    def __repr__(self):
        return repr(self.e)

    def clone(self, e):
        return PName(e, self.name, self.ptree, self.isLeftRec)

    def minLen(self):
        return self.e.minLen()

    def deref(self):
        return self.e.deref()


class PSeq(PTuple):
    def __repr__(self):
        return ' '.join(map(lambda e: self.grouping(e), self.es))

    def minLen(self):
        if not hasattr(self, 'minlen'):
            self.minlen = sum(map(lambda e: e.minLen(), self.es))
        return self.minlen

    @classmethod
    def new(cls, *es):
        if len(es) == 0:
            return EMPTY
        return es[0] if len(es) == 1 else PSeq(*es)

class PAlt(PTuple):
    def __repr__(self):
        return ' | '.join(map(repr, self))


class POre(PTuple):
    def __repr__(self):
        return ' / '.join(map(repr, self))

    def isDict(self):
        for e in self:
            if not isinstance(e, PChar):
                return False
        return True

    def listDict(self):
        dic = [e.text for e in self if isinstance(e, PChar)]
        dic2 = []
        for s in dic:
            if s == '':
                break
            dic2.append(s)
        return dic2

    @classmethod
    def new(cls, *es):
        if len(es) == 0:
            return FAIL
        return es[0] if len(es) == 1 else POre(*es)


class PAnd(PUnary):
    def __repr__(self):
        return '&'+self.grouping(self.e)

    def minLen(self): return 0


class PNot(PUnary):
    def __repr__(self):
        return '!'+self.grouping(self.e)

    def minLen(self): return 0


class PMany(PUnary):
    def __repr__(self):
        return self.grouping(self.e)+'*'

    def minLen(self): return 0


class POneMany(PUnary):
    def __repr__(self):
        return self.grouping(self.e)+'+'


class POption(PUnary):
    def __repr__(self):
        return self.grouping(self.e)+'?'

    def minLen(self): return 0


class PNode(PUnary):
    __slot__ = ['e', 'tag', 'shift']

    def __init__(self, e, tag='', shift=0):
        self.e = e
        self.tag = tag
        self.shift = shift

    def __repr__(self):
        shift = '' if self.shift == 0 else f'/*{self.shift}*/'
        tag = '' if self.tag == '' else f' #{self.tag} '
        return '{' + shift + ' ' + repr(self.e) + tag + '}'


class PEdge(PUnary):
    __slot__ = ['e', 'edge', 'shift']

    def __init__(self, edge, e, shift=0):
        self.e = e
        self.edge = edge
        self.shift = shift

    def __repr__(self):
        shift = '' if self.shift == 0 else f'/*{self.shift}*/'
        return self.edge + shift + ': ' + self.grouping(self.e)


class PFold(PUnary):
    __slot__ = ['e', 'edge', 'tag', 'shift']

    def __init__(self, edge, e, tag='', shift=0):
        self.e = e
        self.edge = edge
        self.tag = tag
        self.shift = shift

    def __repr__(self):
        shift = '' if self.shift == 0 else f'/*{self.shift}*/'
        edge = '^' if self.edge == '' else f'{self.edge}:^'
        tag = '' if self.tag == '' else f' #{self.tag} '
        return '{ ' + edge + shift + ' ' + repr(self.e) + tag + '}'


class PAbs(PUnary):
    __slot__ = ['e']

    def __init__(self, e):
        self.e = e

    def __repr__(self):
        return f'@abs({self.e})'

# Action


class PAction(PUnary):
    __slots__ = ['e', 'func', 'params', 'ptree']

    def __init__(self, e, func, params, ptree=None):
        self.e = e
        self.func = func
        self.params = params
        self.ptree = ptree

    def __repr__(self):
        return f'@{self.func}{self.params}'

    def cname(self):
        return self.func.capitalize()

# CONSTANT
EMPTY = PChar('')
ANY = PAny()
FAIL = PNot(EMPTY)


def isEmpty(pe):
  return pe == EMPTY or isinstance(pe, PChar) and len(pe.text) == 0


def isAny(pe):
  return pe == ANY or isinstance(pe, PAny)


def isSingleCharacter(pe):
        return (isinstance(pe, PChar) and len(pe.text) == 1) or isinstance(pe, PRange) or isinstance(pe, PAny)

# PRange Utilities

def bitsetRange(chars, ranges):
    cs = 0
    for c in chars:
        cs |= 1 << ord(c)
    r = ranges
    while len(r) > 1:
        for c in range(ord(r[0]), ord(r[1])+1):
            cs |= 1 << c
        r = r[2:]
    return cs


def stringfyRange(bits):
    c = 0
    s = None
    p = None
    chars = []
    ranges = []
    while bits > 0:
        if bits & 1 == 1:
            if s is None:
                s = c
                p = c
            elif p + 1 == c:
                p = c
            else:
                _appendRange(s, p, chars, ranges)
                s = c
                p = c
        bits >>= 1
        c += 1
    if s is not None:
        _appendRange(s, p, chars, ranges)
    return ''.join(chars), ''.join(ranges)


def _appendRange(s, p, chars, ranges):
    if s == p:
        chars.append(chr(s))
    elif s+1 == p:
        chars.append(chr(s))
        chars.append(chr(p))
    else:
        ranges.append(chr(s))
        ranges.append(chr(p))

def uniqueRange(chars, ranges):
    bits = bitsetRange(chars, ranges)
    newchars, newranges = stringfyRange(bits)
    checkbits = bitsetRange(chars, ranges)
    assert bits == checkbits
    return newchars, newranges

#
# Visitor
# 

class PVisitor(object):
    def visit(self, pe: PExpr):
        key = self.cname(pe)
        f = getattr(self, key)
        return f(pe)

    def cname(self, pe):
        return pe.cname()


class Nullable(PVisitor):
    def PChar(self, pe): return 1 if len(pe.text) > 0 else 0
    def PAny(self, pe): return 1
    def PRange(self, pe): return 1

    def PRef(self, pe):
        if not hasattr(pe, 'nullable') or pe.nullable is None:
            pe.nullable = 0
            pe.nullable = self.visit(pe.deref())
        return pe.nullable

    def PAnd(self, pe): return 0
    def PNot(self, pe): return 0
    def PMany(self, pe): return 0
    def POneMany(self, pe): return self.visit(pe.e)
    def POption(self, pe): return 0

    def PSeq(self, pe):
        for e in pe:
            if(self.visit(e) == 1):
                return 1
        return 0

    def POre(self, pe):
        for e in pe:
            if(self.visit(e) == 0):
                return 0
        return 1

    def PNode(self, pe): return self.visit(pe.e)
    def PFold(self, pe): return self.visit(pe.e)
    def PEdge(self, pe): return self.visit(pe.e)
    def PAbs(self, pe): return self.visit(pe.e)

    def PAction(self, pe): return self.visit(pe.e)

defaultNullableChecker = Nullable()

def isAlwaysConsumed(pe):
    return defaultNullableChecker.visit(pe)




class First(PVisitor):

    def __init__(self):
        self.memos = {}

    def PChar(self, pe):
        if len(pe.text) == 0:
            return (0, 0)
        return (1 << ord(pe.text[0]), 0)

    def PAny(self, pe):
        return (-1, 0)

    def PRange(self, pe):
        if(pe.neg):
            return (-1, unique_range(pe.chars, pe.ranges))
        return (unique_range(pe.chars, pe.ranges), 0)

    def PRef(self, pe):
        uname = pe.uname()
        if(uname not in self.memos):
            self.memos[uname] = (-1, 0)
            self.memos[uname] = self.visit(pe.deref())
        return self.memos[uname]

    def PNot(self, pe):
        if isSingleCharacter(pe.e):
            pos, neg = self.visit(pe.e)
            return (neg, pos)
        return LeftRef.EMPTYSET

    def PAnd(self, pe): return self.visit(pe.e)
    def PMany(self, pe): return self.visit(pe.e)
    def POneMany(self, pe): return self.visit(pe.e)
    def POption(self, pe): return self.visit(pe.e)

    def PSeq(self, pe):
        pos, neg = 0, 0
        for e in pe:
            pos1, neg1 = self.visit(e)
            pos |= pos1
            neg |= neg1
            if isAlwaysConsumed(e):
                break
        return (pos, neg)

    def POre(self, pe):
        pos, neg = 0, 0
        for e in pe:
            pos1, neg1 = self.visit(e)
            pos |= pos1
            neg |= neg1
        return (pos, neg)

    def PNode(self, pe): return self.visit(pe.e)
    def PFold(self, pe): return self.visit(pe.e)
    def PEdge(self, pe): return self.visit(pe.e)
    def PAbs(self, pe): return self.visit(pe.e)

    def PAction(self, pe): return self.visit(pe.e)






from pegtree.pasm import *


def TPEG(peg):
    pRule(peg, "S", pRange(" \t　", ""))
    pRule(peg, "EOF", pNot(pAny()))
    pRule(peg, "NAME", pSeq2(pRange("_", "AZaz"), pManyRange("_.", "AZaz09")))
    pRule(peg, "UNAME", pOneMany(
        pSeq2(pNotRange("(){}^[]&! \t\r\n　/|*+?.\'\"@:#", ""), pAny())))
    pRule(peg, "Char", pSeq3(pChar("\'"), pNode(pMany(pOre2(pSeq2(pChar("\\"),
                                                                  pAny()), pSeq2(pNotChar("\'"), pAny()))), "Char", 0), pChar("\'")))
    pRule(peg, "DELIM1", pChar("\'\'\'"))
    pRule(peg, "DELIM2", pChar("```"))
    pRule(peg, "Quoted", pSeq2(pChar("\""), pNode(pSeq2(pMany(pOre2(pSeq2(pChar(
        "\\"), pAny()), pSeq2(pNotChar("\""), pAny()))), pChar("\"")), "Quoted", -1)))
    pRule(peg, "Empty", pNode(pEmpty(), "Empty", 0))
    pRule(peg, "Class", pSeq3(pChar("["), pNode(pMany(pOre2(pSeq2(
        pChar("\\"), pAny()), pSeq2(pNotChar("]"), pAny()))), "Class", 0), pChar("]")))
    pRule(peg, "Any", pSeq2(pChar("."), pNode(pEmpty(), "Any", -1)))
    pRule(peg, "Tag", pSeq2(pChar("#"), pSeq2(pSeq2(pNotRange(" \t　\r\n}", ""),
                                                    pAny()), pNode(pMany(pSeq2(pNotRange(" \t　\r\n}", ""), pAny())), "Tag", -1))))
    pRule(peg, "EOL", pOre3(pChar("\n"), pChar("\r\n"), pRef(peg, "EOF")))
    pRule(peg, "Identifier", pNode(
        pOre2(pRef(peg, "NAME"), pRef(peg, "UNAME")), "Name", 0))
    pRule(peg, "COMMENT", pOre2(pSeq3(pChar("/*"), pMany(pSeq2(pNotChar("*/"), pAny())),
                                      pChar("*/")), pSeq2(pChar("//"), pMany(pSeq2(pNot(pRef(peg, "EOL")), pAny())))))
    pRule(peg, "Doc1", pSeq(pChar("\'\'\'"), pManyRange(" \t　", ""), pRef(peg, "EOL"), pNode(pMany(
        pSeq2(pNot(pSeq2(pChar("\'\'\'"), pRef(peg, "EOL"))), pAny())), "Doc", 0), pChar("\'\'\'")))
    pRule(peg, "Doc2", pSeq(pChar("```"), pManyRange(" \t　", ""), pRef(peg, "EOL"), pNode(
        pMany(pSeq2(pNot(pSeq2(pChar("```"), pRef(peg, "EOL"))), pAny())), "Doc", 0), pChar("```")))
    pRule(peg, "Doc0", pNode(pMany(pSeq2(pNot(pRef(peg, "EOL")), pAny())), "Doc", 0))
    pRule(peg, "Ref", pOre2(pRef(peg, "Identifier"), pRef(peg, "Quoted")))
    pRule(peg, "__", pMany(pOre3(pRange(" \t　", ""),
                                 pRange("\r\n", ""), pRef(peg, "COMMENT"))))
    pRule(peg, "_", pMany(pOre2(pRange(" \t　", ""), pRef(peg, "COMMENT"))))
    pRule(peg, "Doc", pOre3(pRef(peg, "Doc1"),
                            pRef(peg, "Doc2"), pRef(peg, "Doc0")))
    pRule(peg, "Names", pNode(pSeq3(pRef(peg, "Identifier"), pRef(peg, "_"), pMany(
        pSeq4(pChar(","), pRef(peg, "_"), pRef(peg, "Identifier"), pRef(peg, "_")))), "", 0))
    pRule(peg, "EOS", pOre2(pSeq2(pRef(peg, "_"), pOneMany(pSeq2(pChar(";"),
                                                                 pRef(peg, "_")))), pOneMany(pSeq2(pRef(peg, "_"), pRef(peg, "EOL")))))
    pRule(peg, "SS", pOre2(pSeq3(pRange(" \t　", ""), pRef(peg, "_"), pNot(pRef(peg, "EOL"))), pSeq3(
        pOneMany(pSeq2(pRef(peg, "_"), pRef(peg, "EOL"))), pRange(" \t　", ""), pRef(peg, "_"))))
    pRule(peg, "Import", pSeq2(pSeq2(pSeq2(pChar("from"), pRange(" \t　", "")), pNode(pSeq3(pRef(peg, "_"), pEdge("name", pOre2(pRef(peg, "Identifier"), pRef(peg, "Char"))),
                                                                                           pOption(pSeq(pRef(peg, "_"), pChar("import"), pRange(" \t　", ""), pRef(peg, "_"), pEdge("names", pRef(peg, "Names"))))), "Import", -9)), pRef(peg, "EOS")))
    pRule(peg, "Example", pSeq2(pSeq2(pSeq2(pChar("example"), pRange(" \t　", "")), pNode(pSeq3(pRef(peg, "_"),
                                                                                               pEdge("names", pRef(peg, "Names")), pEdge("doc", pRef(peg, "Doc"))), "Example", -15)), pRef(peg, "EOS")))
    pRule(peg, "Not", pSeq2(pChar("!"), pNode(
        pEdge("e", pRef(peg, "Predicate")), "Not", 0)))
    pRule(peg, "And", pSeq2(pChar("&"), pNode(
        pEdge("e", pRef(peg, "Predicate")), "And", 0)))
    pRule(peg, "Group", pSeq(pChar("("), pRef(peg, "__"), pOre2(
        pRef(peg, "Expression"), pRef(peg, "Empty")), pRef(peg, "__"), pChar(")")))
    pRule(peg, "Fold", pSeq2(pChar("{"), pNode(pSeq(pRef(peg, "_"), pOre2(pSeq2(pChar("^"), pRef(peg, "__")), pSeq2(pRef(peg, "__"), pEdge("edge", pSeq(pRef(peg, "Identifier"), pChar(":"), pRef(peg, "_"), pChar("^"), pRef(peg, "__"))))), pOption(pSeq2(
        pEdge("tag", pRef(peg, "Tag")), pRef(peg, "__"))), pEdge("e", pOre2(pSeq2(pRef(peg, "Expression"), pRef(peg, "__")), pRef(peg, "Empty"))), pOption(pSeq2(pEdge("tag", pRef(peg, "Tag")), pRef(peg, "__"))), pRef(peg, "__"), pChar("}")), "Fold", -1)))
    pRule(peg, "Node", pSeq2(pChar("{"), pNode(pSeq(pRef(peg, "__"), pOption(pSeq2(pEdge("tag", pRef(peg, "Tag")), pRef(peg, "__"))), pEdge("e", pOre2(pSeq2(pRef(
        peg, "Expression"), pRef(peg, "__")), pRef(peg, "Empty"))), pOption(pSeq2(pEdge("tag", pRef(peg, "Tag")), pRef(peg, "__"))), pRef(peg, "__"), pChar("}")), "Node", -1)))
    pRule(peg, "OldFold", pSeq2(pChar("^"), pNode(pSeq(pRef(peg, "_"), pChar("{"), pRef(peg, "__"), pOption(pSeq2(pEdge("tag", pRef(peg, "Tag")), pRef(peg, "__"))), pEdge("e", pOre2(
        pSeq2(pRef(peg, "Expression"), pRef(peg, "__")), pRef(peg, "Empty"))), pOption(pSeq2(pEdge("tag", pRef(peg, "Tag")), pRef(peg, "__"))), pRef(peg, "__"), pChar("}")), "Fold", -1)))
    pRule(peg, "EdgeFold", pNode(pSeq(pEdge("edge", pRef(peg, "Identifier")), pChar(":"), pRef(peg, "_"), pChar("^"), pRef(peg, "_"), pChar("{"), pRef(peg, "__"), pOption(pSeq2(pEdge("tag", pRef(peg, "Tag")), pRef(
        peg, "__"))), pEdge("e", pOre2(pSeq2(pRef(peg, "Expression"), pRef(peg, "__")), pRef(peg, "Empty"))), pOption(pSeq2(pEdge("tag", pRef(peg, "Tag")), pRef(peg, "__"))), pRef(peg, "__"), pChar("}")), "Fold", 0))
    pRule(peg, "Edge", pNode(pSeq(pEdge("edge", pRef(peg, "Identifier")), pChar(
        ":"), pRef(peg, "_"), pNotChar("^"), pEdge("e", pRef(peg, "Term"))), "Edge", 0))
    pRule(peg, "Func", pSeq2(pChar("@"), pNode(pSeq(pRef(peg, "Identifier"), pChar("("), pRef(peg, "__"), pOre2(pSeq2(pRef(peg, "Expression"), pRef(peg, "_")), pRef(
        peg, "Empty")), pMany(pSeq(pRef(peg, "_"), pChar(","), pRef(peg, "__"), pRef(peg, "Expression"), pRef(peg, "_"))), pRef(peg, "__"), pChar(")")), "Func", -1)))
    pRule(peg, "Suffix", pSeq2(pRef(peg, "Term"), pOption(pOre3(pSeq2(pChar("*"), pFold("e", pEmpty(), "Many", -1)),
                                                                pSeq2(pChar("+"), pFold("e", pEmpty(), "OneMany", -1)), pSeq2(pChar("?"), pFold("e", pEmpty(), "Option", -1))))))
    pRule(peg, "Sequence", pSeq2(pRef(peg, "Predicate"), pOption(
        pFold("", pOneMany(pSeq2(pRef(peg, "SS"), pRef(peg, "Predicate"))), "Seq", 0))))
    pRule(peg, "Choice", pSeq2(pRef(peg, "Sequence"), pOption(pFold("", pOneMany(pSeq4(
        pRef(peg, "__"), pDict("/ ||"), pRef(peg, "_"), pRef(peg, "Sequence"))), "Ore", 0))))
    pRule(peg, "Expression", pSeq2(pRef(peg, "Choice"), pOption(pFold("", pOneMany(pSeq(pRef(
        peg, "__"), pChar("|"), pNotChar("|"), pRef(peg, "_"), pRef(peg, "Choice"))), "Alt", 0))))
    pRule(peg, "Rule", pSeq2(pNode(pSeq(pEdge("name", pOre2(pRef(peg, "Identifier"), pRef(peg, "Quoted"))), pRef(peg, "__"), pDict("= <-"),
                                        pRef(peg, "__"), pOption(pSeq2(pRange("/|", ""), pRef(peg, "__"))), pEdge("e", pRef(peg, "Expression"))), "Rule", 0), pRef(peg, "EOS")))
    pRule(peg, "Statement", pOre3(pRef(peg, "Import"),
                                  pRef(peg, "Example"), pRef(peg, "Rule")))
    pRule(peg, "Source", pNode(pMany(pRef(peg, "Statement")), "Source", 0))
    pRule(peg, "Start", pSeq3(pRef(peg, "__"),
                              pRef(peg, "Source"), pRef(peg, "EOF")))
    pRule(peg, "Term", pOre(pRef(peg, "Group"), pRef(peg, "Char"), pRef(peg, "Class"), pRef(peg, "Any"), pRef(peg, "Fold"), pRef(
        peg, "Node"), pRef(peg, "OldFold"), pRef(peg, "EdgeFold"), pRef(peg, "Edge"), pRef(peg, "Func"), pRef(peg, "Ref")))
    pRule(peg, "Predicate", pOre3(pRef(peg, "Not"),
                                  pRef(peg, "And"), pRef(peg, "Suffix")))
    return peg


TPEGGrammar = TPEG({})
# print(TPEGGrammar)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from pathlib import Path
import pegtree as pg
from pegtree import ParseTree

CJParser = None

def getParser():
  global CJParser
  if CJParser is None:
    CJParser = pg.generate(pg.grammar('cj0.tpeg'))
  return CJParser

VerbForm = {
  'VK5': { 'A': 'か', 'I': 'き', 'T': 'い', 'U': 'く', 'E': 'け', 'O': 'こ', },
  'VS5': { 'A': 'さ', 'I': 'し', 'T': 'し', 'U': 'す', 'E': 'せ', 'O': 'そ', },
  'VT5': { 'A': 'た', 'I': 'ち', 'T': 'っ', 'U': 'つ', 'E': 'て', 'O': 'と', },
  'VN5': { 'A': 'な', 'I': 'に', 'T': 'ん', 'U': 'ぬ', 'E': 'ね', 'O': 'の', },
  'VM5': { 'A': 'ま', 'I': 'み', 'T': 'ん', 'U': 'む', 'E': 'め', 'O': 'も', },
  'VR5': { 'A': 'ら', 'I': 'り', 'T': 'っ', 'U': 'る', 'E': 'れ', 'O': 'ろ', },
  'VW5': { 'A': 'わ', 'I': 'い', 'T': 'っ', 'U': 'う', 'E': 'え', 'O': 'お', },
  'VG5': { 'A': 'が', 'I': 'ぎ', 'T': 'い', 'U': 'ぐ', 'E': 'げ', 'O': 'ご', },
  'VB5': { 'A': 'ば', 'I': 'び', 'T': 'ん', 'U': 'ぶ', 'E': 'べ', 'O': 'ぼ', },
  'V1': { 'A': '', 'I': '', 'T': '', 'U': 'る', 'E': 'れ', 'O': 'よ', },
  'VS': { 'A': 'し', 'I': 'し', 'T': 'し', 'U': 'する', 'E': 'すれ', 'O': 'しよ',},
  'VZ': { 'A': 'じ', 'I': 'じ', 'T': 'じ', 'U': 'ずる', 'E': 'ずれ', 'O': 'ぜよ',},
  'A': { 'A': 'く', 'I': '', 'T': '', 'U': 'い', 'E': 'けれ', 'O': '', },
  'P': { 'A': 'せん', 'I': '', 'T': 'し', 'U': 'す', 'E': '', 'O': '', }, 
}

Mood = {
  'not': ('A', 'な', 'A'),
  'base': ('U', ''),
  'if': ('E', 'ば'),
  'polite': ('I', 'ま', 'P'),
  'noun': ('I', ''),
  'past': ('T', 'た'),
}

def isHira(s):
  return len(s) == 1 and ord('あ') <= ord(s) <= ord('ん')

class CJChunk(object):
  __slots__=['token', 'stem', 'pos', 'extra', 'arguments']

  def __init__(self, stem, pos, extra=None):
    self.token = stem
    self.stem = stem
    self.pos = pos
    self.extra = extra
    self.arguments = None

  def __repr__(self):
    ss = [self.getNormalForm(), self.pos]
    if isinstance(self.extra, list):
      ss.extend(self.extra)
    elif self.extra is not None:
      ss.append(self.extra)
    suffix = self.getSuffix()
    if suffix != '' and self.isNoun():
      ss.append(suffix)
    return repr(tuple(ss))

  def append(self, *values):
    for value in values:
      if value == '': continue
      if self.extra is None:
        self.extra = value
      else:
        if not isinstance(self.extra, list):
          self.extra = [self.extra]
        if value not in self.extra:
          self.extra.append(value)

  def remove(self, *values):
    for value in values:
      if isinstance(self.extra, list):
        if value in self.extra:
          self.extra.remove(value)
        if len(self.extra) == 0:
          self.extra = None
      elif self.extra == value:
        self.extra = None

  def has(self, value):
    if isinstance(self.extra, list):
      return value in self.extra
    return self.extra == value

  def getToken(self):
    return self.token

  def getStem(self):
    return self.stem

  def getSuffix(self):
    return self.token[len(self.stem):]

  def getNormalForm(self, polite=False, pos=None):
    if pos is None:
      pos = self.pos
    if pos in VerbForm:
      return self.stem + VerbForm[pos]['U']
    return self.stem

  def isNoun(self):
    return self.pos.startswith('N')

  def isAdverb(self):
    return self.pos.startswith('NM')

  def isVerb(self):
    return self.pos.startswith('V')

  def isAdj(self):
    return self.pos.startswith('A')

  def isNounPrefix(self):
    if self.isNoun():
      if self.pos == 'NM': return False
      return self.token == self.stem or self.getSuffix().endswith('の')
    if self.isVerb() or self.isAdj():
      return not self.has('@then')

  def isVerbPrefix(self):
    if self.isNoun():
      return not self.getSuffix().endswith('の')
    return self.has('@then')

  def isAdjPrefix(self):
    if self.isNoun():
      return not self.getSuffix().endswith('の')
    return self.has('@then')

  def push(self, chunk):
    if self.arguments is None:
      self.arguments = []
    self.arguments.append(chunk)


class Tokenizer(object):
  def visit(self, node):
    tag = node.getTag()
    if tag == '':
      return self.visit(node[0])
    if tag.endswith('.') or len(node) == 0:
      tag = tag.replace('.', '')
      chunk = CJChunk(node.getToken(), self.pos(tag))      
    else:
      chunk = self.visit(node[0])
      if not tag.startswith('X'):
        chunk.pos = self.pos(tag)
    tags = tag.split('X')
    if len(tags) > 0:
      for meta in tags[1:]:
        if meta.startswith('_'): #X_then
          meta = '@'+meta[1:]
          chunk.remove(meta)
        elif meta != '':
          meta = '@'+meta
          chunk.append(meta)
    return chunk

  def pos(self, tag):
    if tag.startswith('X'): 
      return 'N'
    return tag.split('X')[0]

def readwords(file, suffix=''):
  path = Path(__file__).parent / f'cjdic/{file}'
  with path.open() as f:
    ws = []
    for line in f:
      line = line.replace('\n', suffix)
      ws.append(line)
  return ws


def fit_model(model, w, clz):
  w = w[::-1]
  if w in model:
    clz2 = model[w]
    if clz != clz2:
      model[w] = None
  else:
    model[w] = clz


def fit_word(model, w, i, clz):
  if len(w) == i:
    fit_model(model, w, clz)
  else:
    fit_model(model, w[0:i] + '.', clz)
    if i < 6:
      fit_word(model, w, i+1, clz)


class ReverseModel(object):
  def __init__(self, *dicts):
    self.model = {}
    model = {}
    for clz in dicts:
      ws = readwords(f'{clz}.txt')
      for w in ws:
        w = w[::-1]
        fit_word(model, w, 1, clz)
    for i in range(1, 6):
      for pat in model:
        if len(pat) != i:
          continue
        clz = model[pat]
        if clz is None:
          continue
        if self.predict(pat) is None:
          self.model[pat] = clz

  def predict(self, w, default=None):
    wlen = len(w)
    for p in range(1, 5):
      if p > wlen:
        break
      suffix = w[-p:]
      #print('@', suffix, model.get(suffix, None))
      if suffix in self.model:
        return self.model[suffix]
      if p + 1 > wlen:
        break
      suffix = '.' + suffix
      #print('@@', suffix, model.get(suffix, None))
      if suffix in self.model:
        return self.model[suffix]
    return default


Vt5Model = ReverseModel('VR5', 'VW5', 'VT5')
Vd5Model = ReverseModel('VM5', 'VB5')

CA = "濃篤広弛無怪なこ短狡幼弱紅イ堆聰旨貴好醜白酷汚普善潔憎快暗悪洽著眠太長え尊多懈角硬よ儚永敏穢近眩賢煙軽怖暑高明安固黒遠ぽ煩薄恐偉低畏厚良強狭旧痛鈍古淡粘辛異聡荒憂少凄緩遍熱堅深甘鋭浅豪渋若稚苦恥蒼臭痒羨青吝易乏拙遅い惨温難疎清速懶早幽寒丸赤繁腥細脆円粗佳重酸"
CNA = "恣稀別希妙歪素朧更嫌縦罪酷厭密徒闌端酣初俄切邪変乙主雑暇露粋艷楽急俗顕や雅生純義円重"

CTAIL = {
    #一文字形容詞は以下の文字に限る
    # 'VK5': "繙書乾卷発厭挽画築慄退利戴う届わ拱で囁抱搗急傾欠着履吹扱佩ま敲空弾焼堰抄い割梳び咲ねヤ" +\
    # "靡炊のさ灼頭捌妬が突嘆るふ惹敷あ掻閃なや背曵ゆせ捲砕沸麾碾喚お呟解轟抜聞戦む就し導鳴げ眩驚付" +\
    # "透撒きご耀好描嘯噴劈従ぬ頂撞往ぶ裂覗置た飽浮躓剥懐葺引巻跪塞訊措省溶束泣若焚ら嘶づ漉俯説" +\
    # "どろ行轢除向蠢逝啼磨暴とず拓招咳掃於ば渇は赴め穿破附動哭明如働蠕基釈牽舂て漬点貫疼曳湧歩" +\
    # "播放瞬ぴ輝索舁ぞつざ続涌だ肯蒔結衝研聴か誘徨ひ効呻す欺吐叩布拭斎開じ鋤頷裁碎雪こ響",
    'VT5': "穿落みぎつ過射充待こごた立保経ぶぼ佇育が峙放克分断ま撃満勝なう裁だ建発打滾も伐託擲激か持討毀絶起截",
    'VN5' :"往しい死",
    'VG5': "よ急防ろ戦め訃殺和継拉揺担濯す注扇仰え薙と祝凪瞬し嗣た漱煽稼寛か祈游へ削凌紡ほ告な漕るおゃ騒跨塞炊そ禦剥こ次ぬやはもせ脱接矧喘寿鬻鬱ねわ嗅研扱傾む雪らつ嫁泳さ繋貢の磨",
    # CVW5T = "ぬ唱渫訪被患払犒抗粧負担伝補整報買支結詠諾添よ誓従糾憩あ唄濯逢追疑ゆ迷誘粉る救遣歌" +\
    # "逐恋奪遭喪ま闘向篩ゃ言詛か飼購交揮た養攫願撓謂扱震拭諂随貰戦繕合問通蹲賜請培く酔遵喰ど" +\
    # "構拐伺杓そ嫌希拾の思い纏ろ襲揃狂競わ能祝縫宣会鎧伴使叶争傭厭商衒賄狙掬弔云囲祓舞なぐ倣" +\
    # "這習と候遇蔽匂食笑躇給斎調乞紛幾吸ごも洗失ぶ想ば雇す潤綯漂が沿ふ装贖覆占冀謳やだこ浚窺適償匿" +\
    # "行ょ惑煩ら集呪弄憂奮違謡臭庇お敬慕は曰労"
}

# た行５段活用 活用前の1文字

#
AKA = [
    ('A', 1, '散も善々惨憎濃聡寒ウ易紅ョ白ド粘無荒愛弛色短脆早サ悪イ安ぐ異コぽゃ速怖強み倒少遍尊ガ著ぱ緩堆稚よ低手良ち労疎淡旨憂吝懶高多煩丸辛重ワ固恐狡赤ょ鋭貴懈酸遠豪粗硬細黒円難潔儚篤薄古敏ぼ弱繁臭浅清普腥ロ快暑味温長青佳厚凄幽拙角タく痛け近遅賢畏酷堅渋聰え汚軽煙苦穢醜眠蝿ん暗痒そ恥怪乏幼度深熱狭甘ラ旧鈍ー永洽蒼広偉太羨映'),
    ('VK5', 1, '引頂敲碎て戴哭釈ヤ砕就向画嘶は佩梳解渇戦浮裂働ふ暴焼俯衝轟捲除付割喚で抄驚づ掃歩束湧鋤撒耀跪妬傾吹泣厭効放背響閃結抜牽肯動掻懐咳届説沸抱灼急漬塞研招呻置破於曳乾行築慄繙如従索拓鳴嘆轢省続頭ぬ挽着播斎瞬碾拱曵溶び突拭靡囁基扱ぴ呟往透穿退ひ漉蒔飽惹噴劈剥ざ弾徨聞啼輝空の炊撞開卷赴発逝吐搗訊描嘯捌焚堰点蠢覗附履導欺舂躓麾雪葺頷欠聴書咲疼叩磨舁蠕涌貫敷裁誘巻措利布'),
    ('A', 2, 'いさ|けぶ|とお|好い|かし|頭な|たな|うば|くら|切な|柔か|厳し|馨し|賎し|応し|れし|物し|均し|ら若|むな|しな|めか|にぶ|そば|ふと|がな|った|上な|がし|っこ|ろこ|凛し|慌し|見好|くと|曲し|カし|人し|ちか|ぼろ|ぶか|もた|禍し|斐な|易し|刺し|らた|もし|鹿し|賢し|訳な|荒し|軽し|地好|悪し|味し|騒し|ぼし|つな|だる|ょろ|ざう|どな|さし|はゆ|がる|宜し|いた|どし|気な|そし|暖か|新し|さな|ぢか|もじ|ばし|さま|空し|慮な|なし|しわ|角し|らか|欲し|あま|険し|愛し|清し|妖し|痛し|猛し|くさ|めし|愉し|んど|淋し|のし|重た|ほし|かる|捗し|訝し|可な|危う|題な|むご|陶し|得な|ちな|むた|笑し|らし|あさ|めた|久し|委し|等し|明る|眩し|々し|情な|疾し|びろ|いな|どお|あお|赦な|大き|目好|うし|遽し|詳し|忌し|えご|平た|やわ|らな|芳し|こい|ちろ|すご|怪し|合し|麗し|わか|枉し|つこ|べた|るし|うま|くろ|おぞ|わろ|難し|神し|近し|きな|華し|覚し|ねば|えな|んき|もな|わる|事し|ゲし|ごわ|ねた|とな|賑し|なめ|わい|じか|忙し|しろ|ゆし|ねむ|若し|仰し|侘し|ふる|敢な|香し|まじ|眠た|著し|貧し|ばゆ|るさ|たる|づき|だか|まだ|ふか|目な|軟か|遠し|酷し|継し|哀し|かい|女し|生し|みな|囂し|こわ|映ゆ|男し|しる|斐し|っご|おお|でた|おき|寂し|親し|うた|恥し|由し|うと|せな|けし|すど|疎し|恋し|ぴろ|ひど|うぶ|いし|さと|浅ま|逞し|思し|忝な|床し|てし|あな|際ど|がら|ぬる|くな|香ば|花し|夥し|まな|乏し|美し|ぴど|じろ|うす|烈し|れな|恭し|弱し|ぐろ|ぶせ|危な|福し|びし|まか|けむ|冷た|くし|小さ|あか|でか|こす|好き|ぶな|楽し|えし|雄し|図し|かた|わし|かあ|ずる|のろ|懐こ|まし|卑し|んな|ちい|あら|っき|儀な|おし|から|精し|ばや|厳つ|はや|姦し|ちた|かな|ぐさ|さむ|虚し|憎し|やさ|いろ|労し|悔し|わな|ぐら|まる|こし|やば|なが|細し|ぜな|優し|わゆ|ずき|るど|ずし|げし|せこ|うな|体な|双な|つら|温か|ぎし|とろ|ひろ|たゆ|よわ|ぶと|悲し|白し|初し|っさ|にが|じな|毒し|涼し|がゆ|ぶし|せま|激し|えら|どう|方な|とし|惜し|やし|しこ|珍し|だし|たし|少な|許な|ゆる|むさ|ぎな|くす|あし|ろな|よな|ろし|重し|あや|細か|かゆ|もろ|がた|のう|らう|煙た|喧し|瑞し|やう|懐し|疚し|づら|あつ|わど|たか|しげ|なる|あわ|嬉し|すし|正し|ろう|けな|愛な|げな|ざと|苦し|やす'),
    ('VK5', 2, 'なず|なげ|ちゆ|春め|りむ|くだ|きさ|なな|めゆ|浮つ|軋め|ふぶ|ミつ|艷め|ブつ|ずつ|えつ|かま|しお|心ゆ|りぞ|えお|はた|ラつ|たぶ|てお|粘つ|わむ|りさ|いだ|げつ|さや|だめ|タつ|わめ|うぞ|夏め|ろゆ|似つ|たむ|みが|えす|傷つ|時め|芽ぶ|のめ|ちつ|ため|せつ|とつ|めお|仄め|やぶ|そむ|ぼや|ほど|まめ|ゆわ|すず|えが|しだ|さめ|ぼつ|うず|っつ|らつ|いお|けつ|もむ|ろめ|ざめ|みむ|よめ|秋め|娜め|なつ|にま|りお|ゆが|いぶ|ぞめ|犇め|につ|さば|くつ|煌め|ぶつ|きず|色め|りま|しご|くじ|ばつ|すだ|らめ|あが|がつ|めぶ|ある|息ま|おむ|ろつ|ひら|皮む|渦ま|つむ|かわ|つめ|えき|そぶ|ずま|もど|れつ|ゃつ|かず|わつ|さつ|はぶ|てつ|とか|びゆ|でむ|しめ|うご|囁や|うめ|るめ|もと|寝つ|やめ|きま|つつ|ぎゆ|たた|しず|ぶや|ただ|のぞ|はじ|らま|れき|きめ|りつ|わぶ|いつ|きお|びつ|みつ|ばた|ごめ|ぱた|たつ|もが|がや|あば|ぼめ'),
    ('A', 3, 'まりな|でっか|りりし|たっと|限りな|頼りな|あたら|並びな|らびな|がめつ|ちくど|さみし|だざむ|どぎつ|あくど|よりな|きまず|りくど|わりな|いとど|いかつ|足りな|あまね|弔りし|気まず|たりな|ぎりな'),
    ('VK5', 3, 'だきつ|のびな|ちりし|ざまず|ひっか|はたら|すりな|いきつ|なきつ|しまね|きとど|ふりし|まごつ|とどろ|やきつ|くみし|こぎつ|ゆきつ|泣きつ|抱きつ|せびな|むかつ|まきつ|引っか|じめつ|つまず|きくど|行きつ|よぎつ|あざむ|ほっと|こまね'),
    ('A', 4, 'ろおどろ')
]
AWA = [
    ('A', 1, '散善々惨憎濃聡寒ウ易紅ョ白せド粘無荒愛弛色短脆早サ悪イ若安異コぽ速怖強み倒少遍尊ガ著ぱう緩堆稚ぞ低手良ちげ疎淡旨吝つ懶高多丸辛重ワ固恐狡ず赤鋭貴懈酸遠豪粗硬細し黒円難潔儚明篤薄古敏ぼ弱じ繁浅清普腥ロ快暑む味温長青佳厚凄幽拙角タ痛け近遅賢畏酷堅渋ね聰え汚眩軽煙苦さ穢醜眠蝿ん暗痒恥め怪乏幼度深熱狭甘好ラ旧き鈍ー永洽蒼広偉太羨映'),
    ('VW5', 1, '通嫌拾衒鎧攫撓払向宣遣伺這紛は弄漂襲狂唱囲賜戦纏篩遇冀祝ふ笑償迷思狙謳整養闘負担逢候失購唄抗拐奪蹲使謂集伝厭祓請結曰占揃誓贖随呪潤雇救揮訪躇被装能交患追合問震遭恋喰行幾従会犒乞支想窺ぬ詠粧斎食調補違培拭叶掬扱酔諂粉願慕添吸渫糾言覆匿買飼構争の弔喪憩適縫浚繕諾沿匂敬疑競商庇杓伴倣報云希奮濯歌貰賄洗詛舞習遵逐傭綯謡誘給蔽惑'),
    ('A', 2, 'けぶ|とお|えぐ|好い|頭な|たな|切な|づよ|どよ|水臭|柔か|別臭|散臭|むな|めか|にぶ|そば|ふと|った|上な|ひく|っこ|ろこ|とど|人臭|黴臭|くと|ぼろ|気臭|斐な|ぶか|もた|らた|び臭|乳臭|やす|つな|だる|ょろ|どな|はゆ|がる|いた|磯臭|つよ|気な|暖か|さな|ぢか|さま|慮な|ろよ|しわ|鹿臭|あま|小煩|しぶ|っと|そ臭|んど|重た|かる|こよ|可な|むご|題な|得な|ちな|むた|りな|めた|土臭|明る|げ臭|情な|びろ|いな|どお|あお|っか|赦な|えご|平た|やわ|こい|ちろ|すご|わか|べた|つこ|向臭|うま|落臭|わろ|照臭|地よ|びな|ねば|えな|わる|ごわ|ぼそ|ねた|わい|じか|しろ|くど|ぬく|敢な|眠た|ばゆ|たる|心憂|な臭|た臭|だか|タ臭|まだ|ふか|目な|軟か|ほそ|かい|かく|みな|こわ|映ゆ|ぎよ|しる|柿臭|っご|でた|うと|せな|すど|ぴろ|ひど|うぶ|さと|浅ま|忝な|さく|焦臭|暮臭|あな|際ど|がら|ぬる|くな|香ば|まな|じろ|金臭|ぴど|うす|とも|れな|ぐろ|危な|ちよ|めよ|どろ|まか|冷た|あか|きよ|でか|こす|ぶな|かた|倒臭|香臭|かあ|ずる|懐こ|の憂|んな|ちい|古臭|みよ|儀な|ばや|はや|ちた|労労|いろ|味よ|泥臭|わな|ぐら|まる|れ臭|やば|なが|ぜな|わゆ|るど|せこ|うな|体な|口煩|双な|温か|とろ|よわ|たゆ|ぶと|にが|にく|がゆ|程よ|せま|好よ|青臭|えら|方な|しこ|生臭|少な|許な|ゆる|くす|物憂|ろな|よな|あや|細か|かゆ|もろ|がた|煙た|見よ|づら|わど|なる|あわ|けな|愛な|げな|ざと|訳な'),
    ('VW5', 2, 'じあ|い煩|りあ|荷な|さよ|けあ|そぐ|しあ|きそ|いら|巣く|やと|ぎわ|にな|かま|みま|らそ|そお|けお|つろ|さら|りそ|映ろ|休ら|つど|もら|かば|むく|むら|計ら|まが|安ら|びか|そろ|つだ|かこ|びあ|むか|きら|語ら|ちが|ずら|にあ|きあ|ちあ|いわ|しら|ねら|べな|なら|づか|れそ|はら|ざな|れあ|たま|向か|きわ|ぎら|てが|てら|ねあ|であ|よろ|まご|住ま|つが|みあ|まよ|すら|まど|よそ|さそ|こな|あた|じわ|のい|出あ|かが|腹ば|しま|めら|めあ|いこ|すく|ぱら|じら|にお|まと|振る|せお|るお|ゃく|たが|行な|ぐな|いあ|らば|やま|喰ら|わら|じま|ねが|味わ|だよ|ちそ|いと|くま|るま|した|じゃ|逆ら|くば|らが|くる|うろ|すま|ちま|ゆた|損な|食ら|立あ|きお|つか|ぬぐ|おぶ|たぐ|もや|きか|つた|賑わ'),
    ('A', 3, 'はかな|よしな|かとな|まらな|しがな|もとな|わしな|ぢきな|かうば|忙しな|ったか|下らな|堪らな|まから|っかな|つもな|うもな|いじな|ずおお|うがな|れおお|おから|てあら|すかな|きしょ|好かな|よぎな|てしな|軟らか|ずたか|たとな|だらな|さがな|あたら|すから|っちゃ|でもな|わらか|たらな|なかよ|ごとな|つがな|りおお|ならな|らうた|もくろ|らしな|っから|ばひろ|柔らか|つかな|るぎな|ておも|っくろ|じきな'),
    ('VW5', 3, 'おとな|似かよ|あがな|うしな|つくろ|にかよ|かたら|はから|からか|へつら|のおも|やしな|うらな|まじな|あきな|んくら|つちか|まかな|げつら|さから|おぎな|面くら'),
    ('A', 4, 'っともな|あたたか'),
    ('VW5', 4, 'いともな')
]


def infer(dataset, w, default_pos='A'):

  for pos, size, pat in dataset:
      if len(w) >= size and w[-size:] in pat:
        return pos
  return default_pos


def setpos(c: CJChunk, pos):
  c.pos = pos
  return c

# check_nai(chunk)

def check_nai(chunk):
  CNAI = 'わらいみけらいかがんむげぜらびちま'
  WNAI = 'おもいがけ|ちがい|あじけ|危なげ|惜しみ|いわけ|なにげ|おとなげ|ほい|おおけ|なさけ|のっぴきなら|もうしわけ|つまら|くだら|そぐわ|ままなら|まちがい|おしみ|あっけ|居たたまら|たわい|思いがけ|いけ好か|しようが|違い|間違い|びん|情け|いけすか|つつが|やむ|味け|しょうが|大人げ|がんぜ|くちさが|すげ|たまら|あい|はか|下ら|しが|ならび|並び|ぎこち|じょさい|つまん|もったい|しどけ|ふがい|おほけ|おもん|あどけ|すけ|かたじけ|居た堪ら|おぼつか|たあい|済ま|ぎごち|さりげ|そっけ|にげ|口さが|ものたら|おっか|いたたまら|申しわけ|あたじけ|いとけ|せん|しょざい'
  if chunk.pos.startswith('V') and chunk.has('@not'):
    stem = chunk.getStem()
    if len(stem) > 1:
      if stem[-1] not in CNAI or stem in WNAI:
        nstem = chunk.stem + 'な'
        if chunk.token.startswith(nstem):
          chunk.stem = nstem
          chunk.append(f'@verb({chunk.pos})')
          chunk.pos = 'A'
          chunk.remove('@not', '@can', '@passive', '@make')
          return chunk
  return chunk

#NNO = 'その,あの,どの,この'.split(',')

def normalize(chunk: CJChunk):
  w = chunk.stem
  if chunk.pos == 'AN':
    if len(w) == 2 and w[0] not in CNA:
      chunk.pos = 'N'
    return chunk
  if chunk.pos in CTAIL and w[-1] not in CTAIL[chunk.pos] :
    chunk.stem = chunk.token
    chunk.pos = 'N'
    return chunk
  if chunk.pos == 'AVK5':
    chunk.pos = infer(AKA, w)
    if chunk.pos == 'VK5':
      chunk.remove('@then')
    return chunk
  if chunk.pos == 'Vd5':
    chunk.pos = Vd5Model.predict(w, 'VM5')
    return chunk
  if chunk.pos == 'Vt5':
    chunk.pos = Vt5Model.predict(w, 'VR5')
    return chunk
  return check_nai(chunk)


def isPrefix(c: CJChunk, pos):
  return c.pos == pos and c.extra is None

def isComposable(c: CJChunk):
  pos = c.pos
  return pos == 'N' or pos.startswith('V') or pos.startswith('A')

def isToken(c: CJChunk, pos, token):
  return c.pos == pos and c.token == token

def concat1(c: CJChunk, c2: CJChunk):
  c.token = c.token + c2.token
  return c

def concat2(c: CJChunk, c2: CJChunk):
  c2.token = c.token + c2.token
  return c2

def concat(c: CJChunk, c2: CJChunk):
  #ルール1 の前に、名詞 
  #格好いい[('格好い', 'A'), ('い', 'N')]
  if isToken(c2, 'N', 'い'):
    c.pos = 'A'
    c.append(f'@suffix({c2.token})')
    return concat1(c, c2)
  if isToken(c2, 'N', 'り'):
    c2.stem = c.token + c2.stem
    c2.append(f'@suffix({c2.token})')
    return concat2(c, c2)
  # ルール1. 名詞NVは接続される
  # [('もの', 'N'), ('ぐるわしい', 'A')]
  if isPrefix(c, 'NV') and isComposable(c2):
      #print('@concat', c.token, c2.token)
      c2.stem = c.token + c2.stem
      c2.append(f'@prefix({c.token})')
      return concat2(c, c2)
  return None

def push_argument(chunk, prefix):
  if chunk.arguments is not None:
    for a in chunk.arguments[::-1]:
      if push_argument(a, prefix):
        return True
  if chunk.isNoun() and prefix.isNounPrefix():
    print('pushN', repr(prefix), repr(chunk))
    chunk.push(prefix)
    return True
  if chunk.isVerb() and prefix.isVerbPrefix():
    print('pushV', repr(prefix), repr(chunk))
    chunk.push(prefix)
    return True
  if chunk.isAdj() and prefix.isAdjPrefix():
    print('pushA', repr(prefix), repr(chunk))
    chunk.push(prefix)
    return True
  print('!push', repr(prefix), repr(chunk))
  return False

def stringfy_arguments(chunk):
  if chunk.arguments is None:
    return repr(chunk)
  ss = []
  for argument in chunk.arguments[::-1]:
    ss.append(stringfy_arguments(argument))
  ss.append(repr(chunk))
  return '[' + ' '.join(ss) + ']'

def make_arguments(chunks):
  if len(chunks) == 0:
    return
  chunks = chunks[::-1]
  target = chunks[0]
  for prefix in chunks[1:]:
    push_argument(target, prefix)
  print('@@', stringfy_arguments(target))

def tokenize(text, parser = None):
  if isinstance(text, ParseTree):
    tree = text
  else:
    if parser is None:
      parser = getParser()
    try:
      tree = getParser()(text)
    except RecursionError:
      print('FIXME(RecursionError)', text)
      return []
  tokenizer = Tokenizer()
  chunks = []
  for node in tree.getSubNodes():
    s = node.getToken()
    chunk = tokenizer.visit(node)
    chunk.token = s
    chunk = normalize(chunk)
    if len(chunks)>0:
      cat = concat(chunks[-1], chunk)
      if cat is not None: chunks[-1] = cat; continue
    chunks.append(chunk)
  #make_arguments(chunks)
  return chunks

def segment(s: str, sep='/', parser = None):
  chunks = tokenize(s, parser)
  return sep.join([x.token for x in chunks])

#print(tokenize('望遠鏡で子犬が泳ぐのを見た'))
#print(tokenize('望遠鏡で{{すべての子犬が泳ぐのを}}見た'))



class Subnode(list):
    pass

def nocolor(s): 
    return s

class ParseTree(list):
    tag: str
    source: str
    spos: int
    epos: int
    urn: str
    child: Subnode

    def init(self, tag, source, spos=0, epos=None, urn='(unknown source)'):
        self.tag = tag
        self.source = source
        self.spos = spos
        self.epos = epos if epos is not None else len(source)
        self.urn = urn
        self.child = None

    def getTag(self):
        return self.tag

    def isSyntaxError(self):
        return self.tag == 'err'

    def __eq__(self, tag):
        return self.tag == tag if isinstance(tag, str) else self is tag

    def getPosition(self):
        return rowcol(self.urn, self.source, self.spos)

    def getEndPosition(self):
        return rowcol(self.urn, self.source, self.epos)

    def decode(self):
        source, spos, epos = self.source, self.spos, self.epos
        LF = b'\n' if isinstance(source, bytes) else '\n'
        rows = source[:spos + (1 if len(source) > spos else 0)]
        rows = rows.split(LF)
        linenum, column = len(rows), len(rows[-1])-1
        begin = source.rfind(LF, 0, spos) + 1
        # print('@', spos, begin, source)
        end = source.find(LF, spos)
        # print('@', spos, begin, source)
        if end == -1:
            end = len(source)
        # print('@[', begin, spos, end, ']', epos)
        line = source[begin:end]  # .replace('\t', '   ')
        mark = []
        endcolumn = column + (epos - spos)
        for i, c in enumerate(line):
            if column <= i and i <= endcolumn:
                mark.append('^' if ord(c) < 256 else '^^')
            else:
                mark.append(' ' if ord(c) < 256 else '  ')
        mark = ''.join(mark)
        return (self.urn, spos, linenum, column, line, mark)

    def message(self, msg='Syntax Error'):
        urn, pos, linenum, cols, line, mark = self.decode()
        return '{} ({}:{}:{}+{})\n{}\n{}'.format(msg, urn, linenum, cols, pos, line, mark)

    # def subs(self):
    #     es = []
    #     for i, child in enumerate(self):
    #         es.append((child.spos, '', child))
    #     for key in self.dict:
    #         v = self.dict[key]
    #         if isinstance(v, ParseTree):
    #             es.append((v.spos, key, v))
    #     es.sort()
    #     return [(x[1], x[2]) for x in es]

    def isEmpty(self):
        return self.tag == 'empty'

    def newEmpty(self):
        return ParseTree('empty', self.source, self.epos, self.epos, self.urn)

    def set(self, key, tree):
        assert isinstance(tree, ParseTree)
        if self.child is None:
            self.child = Subnode()
        if key == '':
            self.child.append(tree)
        else:
            setattr(self.child, key, tree)
    
    def getNodeSize(self):
        return 0 if self.child is None else len(self.child)
    
    def __len__(self):
        return 0 if self.child is None else len(self.child)

    def getSubNodes(self):
        return [] if self.child is None else list(self.child)

    def __contains__(self, key):
        if self.child is not None:
            if isinstance(key, str):
                return hasattr(self.child, key)
            elif isinstance(key, int):
                return 0 <= key < len(self.child) or -len(self.child) < key <= -1:
        return False
    
    def has(self, key):
        return self.__contains__(key)

    def __getitem__(self, key):
        if self.child is not None:
            if isinstance(key, str):
                return getattr(self.child, key)
            else:
                return self.child[key]
        return self.NoChild(key)

    def get(self, key):
        return self.__getitem__(key)

    def find(self, *key):
        if self.child is not None:
            for key in keys:
                if key in self:
                    return self[key]
        return self.NoChild(key)
    
    def keys(self):
        if self.child is not None:
            return self.child.__dict__.keys()

    def getToken(self, key=None, default_token=''):
        if key is None:
            s = self.source[self.spos:self.epos]
            return s
        return self[key].getToken() if key in self else default_token

    # def substring(self, start=None, end=None):
    #     if start is None:
    #         if end is None:
    #             return self.getToken()
    #         s = self.source[end.epos:self.epos]
    #     else:
    #         if end is None:
    #             s = self.source[self.spos: start.spos]
    #         else:
    #             s = self.source[start.epos:end.spos]
    #     return s.decode('utf-8') if isinstance(s, bytes) else s

    def str(self):
        s = self.source[self.spos:self.epos]
        return str(s)

    def allEdges(self):
        if self.child is not None:
            edges = []
            for n, tree in enumerate(self.child):
                edges.append((tree.spos, n, tree))
            for edge in self.child.__dict__:
                tree =  self.child.__dict__[edge]
                edges.append((tree.spos, edge, tree))
            edges.sort()
            return edges
        return []

    def repr(self):
        if self.isSyntaxError():
            return self.message('Syntax Error')
        sb = []
        self.strOut(sb, indent='', tab='')
        return "".join(sb)

    def strOut(self, sb, indent='\n  ', tab='  ', prefix='', tag=nocolor, edge=nocolor, token=nocolor):
        sb.append(indent + prefix + "[" + tag(f'#{self.getTag()} '))
        edges = self.allEdges()
        if len(edges) > 0:
            nextindent = indent + tab
            for _, label, child in edges:
                prefix = edge(label) + ': ' if isinstance(label, str) else ''
                child.strOut(sb, nextindent, tab, prefix, tag, edge, token)
            sb.append(indent + "]")
        else:
            sb.append(token(repr(str(self))))
            sb.append("]")
    
    def dump(self, indent='\n', tab='  ', tag=nocolor, edge=nocolor, token=nocolor):
        if self.isSyntaxError():
            print(self.message('Syntax Error'))
        else:
            sb = []
            self.strOut(sb, indent, tab, '', tag, edge, token)
            print("".join(sb))

def rowcol(urn, source, spos):
    source = source[:spos + (1 if len(source) > spos else 0)]
    rows = source.split(b'\n' if isinstance(source, bytes) else '\n')
    return len(rows), len(rows[-1])-1


if __name__ == '__main__':
    t = ParseTree


import os

# Option
DebugFlag = 'DEBUG' in os.environ
VerboseFlag = 'VERBOSE' in os.environ
istty = True


def DEBUG(*x):
    if DebugFlag:
        print('DEBUG', *x)


def VERBOSE(*x):
    if DebugFlag:
        print(*x)


COLOR = {
    "Black": '0;30', "DarkGray": '1;30',
    "Red": '0;31', "LightRed": '1;31',
    "Green": '0;32', "LightGreen": '1;32',
    "Orange": '0;33', "Yellow": '1;33',
    "Blue": '0;34', "LightBlue": '1;34',
    "Purple": '0;35', "LightPurple": '1;35',
    "Cyan": '0;36', "LightCyan": '1;36',
    "LightGray": '0;37', "White": '1;37',
}

class Terminal(object):
    istty: bool
    isverbose: bool
    def __init__(self):
        self.istty = True
        self.isverbose = False

    def bold(self, s):
        return '\033[1m' + str(s) + '\033[0m' if self.istty else str(s)

    def color(self, c, s):
        return '\033[{}m{}\033[0m'.format(COLOR[c], str(s)) + '' if self.istty else str(s)

    def perror(self, ptree, msg='syntax error'):
        print(self.color('Red', '[error]'), ptree.message(msg))

    def pwarn(self, ptree, msg):
        if self.isverbose:
            print(self.color('Purple', '[warning]'), ptree.message(msg))

    def pinfo(self, ptree, msg):
        if self.isverbose:
            print(self.color('Cyan', '[info]'), ptree.message(msg))

DefaultConsole = Terminal()




#!/usr/local/bin/python
from pathlib import Path
import functools
import subprocess
# import readline
import time
import sys
import os
import importlib
import pegtree
import pegtree.treeconv as treeconv
from pegtree.terminal import DefaultConsole

bold = DefaultConsole.bold
color = DefaultConsole.color

'''
istty = True


def bold(s):
    return '\033[1m' + str(s) + '\033[0m' if istty else str(s)


COLOR = {
    "Black": '0;30', "DarkGray": '1;30',
    "Red": '0;31', "LightRed": '1;31',
    "Green": '0;32', "LightGreen": '1;32',
    "Orange": '0;33', "Yellow": '1;33',
    "Blue": '0;34', "LightBlue": '1;34',
    "Purple": '0;35', "LightPurple": '1;35',
    "Cyan": '0;36', "LightCyan": '1;36',
    "LightGray": '0;37', "White": '1;37',
}




def showing(pos, msg):
    if pos is None:
        print(msg)
    else:
        print(pos.showing(msg))

def log(type, pos, *msg):
    msg = ' '.join(map(str, msg))
    if type.startswith('err'):
        showing(pos, color('Red', '[error] ') + str(msg))
    elif type.startswith('warn'):
        showing(pos, color('Orange', '[warning] ') + str(msg))
    elif type.startswith('info') or type.startswith('notice'):
        showing(pos, color('Cyan', '[info] ') + str(msg))
    else:
        showing(pos, str(msg))
'''

def version():
    print(DefaultConsole.bold('PEGTree - A PEG Parser Generator with Tree Annotation'))


def read_inputs(a):
    path = Path(a)
    if path.exists():
        f = path.open()
        data = f.read()
        f.close()
        return data
    else:
        return a


def readlines(prompt):
    s = input(prompt)
    if s != '':
        return s
    else:
        l = []
        while True:
            prev = s
            s = input()
            l.append(s)
            if prev == '' and s == '':
                break
        return '\n'.join(l)

def parse_options(argv):


    options = {
        '-v': ('verbose', True),
        '-verbose': ('verbose', True),
        '-g': ('grammar', None),
        '--grammar': ('grammar', None),
        '-s': ('start', None),
        '--start': ('start', None),
        '-e': ('expression', None),
        '--expr': ('expression', None),
        '-f': ('format', None),
        '--format': ('format', None),
        '-O0': ('-O', 0),
        '-O1': ('-O', 1),
        '-O2': ('-O', 2),
        '-O2': ('-O', 3),
    }
    def parse_each(a, d):
        first = a[0]
        if first.startswith('-'):
            if first in options:
                key, value = options[first]
                if value is None:
                    if len(a) > 1:
                        d[key] = a[1]
                        return a[2:]
                else:
                    d[key] = value
                    return a[1:]
            #d['inputs'].append(a)
            raise CommandUsageError
        else:
            d['inputs'].append(a[0])
            return a[1:]

    d = {'inputs': [], '-O': 2, 'verbose': False}
    while len(argv) > 0:
        argv = parse_each(argv, d)
    #print('OPTION', d)
    if d['verbose'] :
        DefaultConsole.isverbose = True
    return d


class CommandUsageError(Exception):
    pass


def usage():
    print(bold('PEGTree - A PEG Parser Generator with Tree Annotation'))
    print("Usage: pegtree <command> options inputs")
    print("  -g | --grammar <file>      specify a grammar file")
    print("  -e | --expr <expression>   specify a parsing expression")
    print("  -s | --start <NAME>        specify a starting rule")
    print("  -f | --format <file>       specify an output format")
    print()

    print("Example:")
    print("  pegtree parse -g math.tpeg <inputs>")
    print("  pegtree example -g math.tpeg <inputs>")
    print("  pegtree pasm -g math.tpeg")
    print("  pegtree update")
    print()

    print("The most commonly used pegtree commands are:")
    print(" parse      run a generated parser")
    print(" example    test all examples")
    print(" pasm       generate a pasm combinator function")
    print(" list       all sample grammars")
    print(" update     update pegtree (via pip)")

showingTPEG = False

def load_grammar(options, default=None):
    global showingTPEG
    expr = options.get('expression', None)
    if expr is not None:
        grammar = 'A = ' + expr
        options['urn'] = f'(-e {repr(expr)})'
        return pegtree.grammar(grammar, **options)
    file = options.get('grammar', default)
    if file is None:        
        print('Enter a TPEG grammar')
        sb = []
        try:
            while True:
                s = input()
                if s == '' or s is None:
                    break
                sb.append(s)
        except:
            pass
        data = '\n'.join(sb)
        options['urn'] = '(stdin)'
        showingTPEG = False
        return pegtree.grammar(data, **options)
    if file == 'stdin.tpeg':
        data = sys.stdin.read()
        options['urn'] = file
        return pegtree.grammar(data, **options)
    options['urn'] = file
    return pegtree.grammar(file, **options)


def generator(options):
    #if 'parser' in options:
    #    m = importlib.import_module(options['parser'])
    #    return m.generate
    return pegtree.generate


def getstart(peg, options):
    if 'start' in options:
        return options['start']
    return peg.start()


def colorTree(t):
    if t.isSyntaxError():
        return t.message(color('Red', 'Syntax Error'))
    else:
        # unconsumed = ''
        # if epos < len(t.inputs_):
        #     unconsumed = ' + ' + color('Purple', t.inputs_[epos:])
        sb = []
        t.strOut(sb, token=lambda x: color('Blue', x),
                 tag=lambda x: color('Cyan', x))
        return "".join(sb)

# parse command

def sample(options):
    files = os.listdir(Path(__file__).parent / 'grammar')
    files.sort()
    for file in files:
        if file.endswith('.tpeg'):
            print(file)


def tpeg(options):
    peg = load_grammar(options)
    print(peg)
    if '@@example' in peg:
        print()
        prefix = color('Blue', 'example')
        quote = color("Red", "'''")
        for testcase in peg['@@example']:
            name, doc = testcase
            name = bold(name)
            doc = doc.getToken()
            if doc.find('\n') > 0:
                print(prefix, name, quote)
                print(doc, quote, sep='\n')
            else:
                print(prefix, name, doc)


def peg(options):
    options['isPurePEG'] = True
    peg = load_grammar(options)
    print(peg)


def optimize(options):
    from pegtree.optimizer import prepare
    peg = load_grammar(options)
    start, refs, rules, memos = prepare(peg)
    for ref in refs:
        uname = ref.uname(peg)
        print(uname, '=', rules[uname])
    print('memo:', ' '.join(memos))


def parse(options, conv=None):
    peg = load_grammar(options)
    parser = generator(options)(peg, **options)
    inputs = options['inputs']
    tdump = treeconv.treedump(options, colorTree)
    if len(inputs) == 0:  # Interactive Mode
        try:
            start = getstart(peg, options)
            while True:
                s = readlines(color('Blue', start) + bold(' <<< '))
                tree = parser(s, urn='(stdin)')
                print(tdump(tree))
        except (EOFError, KeyboardInterrupt):
            pass
    elif len(inputs) == 1:
        colorTree(read_inputs(inputs[0]))
    else:
        for file in options['inputs']:
            st = time.time()
            t = parser(read_inputs(file))
            et = time.time()
            print(file, (et - st) * 1000.0, "[ms]:", t.tag)


def example(options):
    peg = load_grammar(options)
    if '@@example' not in peg:
        return
    parsers = {}
    for testcase in peg['@@example']:
        name, doc = testcase
        if not name in peg:
            continue
        if not name in parsers:
            parsers[name] = generator(options)(peg, start=name)
        res = parsers[name](doc.inputs_, doc.urn_, doc.spos_, doc.epos_)
        # print()
        ok = doc.inputs_[doc.spos_:res.epos_]
        fail = doc.inputs_[res.epos_:doc.epos_]
        print(bold(f'parsing {name}'), color(
            'Green', f'{ok}')+color('Red', f'{fail}'), bold('=> '), end='')
        print(colorTree(res))


def dumpError(lines, line, s):
    errs = 0
    for t in s:
        cur = str(t)
        if t.isSyntaxError():
            errs = 1
            s = max(0, t.spos_ - 10)
            prev = t.inputs_[s:t.spos_]
            # print(line)
            print(lines, color('Green', f'{prev}') + color('Red', f'{cur}'))
    return errs


def test(options):
    peg = load_grammar(options)
    parser = generator(options)(peg, **options)
    inputs = options['inputs']
    st = time.time()
    lines = 0
    fail = 0
    try:
        for file in options['inputs']:
            with open(file) as f:
                for line in f:
                    lines += 1
                    try:
                        t = parser(line)
                        fail += dumpError(lines, line, t)
                    except RecursionError:
                        print(color('Red', line))
                        fail += 1
    except KeyboardInterrupt:
        pass
    et = time.time()
    if lines > 0:
        print(f'{fail}/{lines} {fail/lines} {(et - st) * 1000.0} ms')


def pasm(options):
    from pegtree.nezcc import parsec
    peg = load_grammar(options)
    parsec(peg, **options)


def pasmcc(options):
    from pegtree.nezcc import parsec
    peg = load_grammar(options)
    parsec(peg, **options)


def cjtoken(options, conv=None):
    import pegtree.cj as cj
    peg = load_grammar(options)
    parser = generator(options)(peg, **options)
    inputs = options['inputs']
    for file in options['inputs']:
        with open(file) as f:
            for line in f.readlines():
                if line.startswith('#'): continue 
                line = line.replace('\n', '')
                print(line)
                tree = parser(line)
                print(repr(tree))
                for token in cj.tokenize(tree):
                    print(repr(token))
                print()

def update(options):
    try:
        # pip3 install -U git+https://github.com/KuramitsuLab/pegpy.git
        subprocess.check_call(
            ['pip3', 'install', '-U', 'pegtree'])
    except:
        pass


def update_beta(options):
    try:
        # pip3 install -U git+https://github.com/KuramitsuLab/pegtree.git
        subprocess.check_call(
            ['pip3', 'install', '-U', 'git+https://github.com/KuramitsuLab/pegtree.git'])
    except:
        pass

def main(argv=sys.argv):
    names = globals()
    if len(argv) > 1:
        cmd = argv[1]
        options = parse_options(argv[2:])
        cs = cmd.split('.')
        if len(cs) == 2:
            cmd = cs[0]
            options['ext'] = cs[1]
        if cmd in names:
            names[cmd](options)
            return
    usage()



if __name__ == "__main__":
    main(sys.argv)



import pegtree as pg
import pegtree.graph as graph
from IPython.core.magic import register_cell_magic

def parse_example(peg, line):
    if '@@example' in peg and len(peg['@@example']) > 0:
        name, doc = peg['@@example'][-1]
        parser = pg.generate(peg, start=str(name))
        return parser(str(doc))
    else:
        parser = pg.generate(peg)
        return parser(line)

@register_cell_magic
def TPEG(line, src):
    peg = pg.grammar(src)
    tree = parse_example(peg, line)
    if tree.isSyntaxError():
        print(repr(tree))
    else:
        print(repr(tree))
        return graph.Viz(tree)

@register_cell_magic
def pegtree(line, src):
    return TPEG(line, src)

@register_cell_magic
def match(line, src):
    peg = pg.grammar(src)
    if '@error' not in peg:
        parser = pg.generate(peg)
        res = parser(line)
        print(repr(res))


@register_cell_magic
def parse(line, src):
    peg = pg.grammar(src)
    if '@error' not in peg:
        parser = pg.generate(peg)
        res = parser(line)
        if res.isSyntaxError():
            print(repr(res))
        else:
            print(repr(res))
            return graph.Viz(res) if res is not None else None

